# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from capellambse import MelodyModel


def test_PhysicalPath_has_ordered_list_of_involved_items(model: MelodyModel):
    expected = [
        "544549d6-2aa4-44c2-b2ae-a86302f48e62",
        "42ee9e89-d445-45a2-8280-028d4fb1038d",
        "ea8cd402-3d9a-469a-a2ce-2bc1252c3a01",
        "3078ec08-956a-4c61-87ed-0143d1d66715",
        "baa3047c-9cb4-40a7-9b67-b9b5f76fd2ee",
    ]

    path = model.pa.all_physical_paths.by_name("card1 - card2 connection")

    actual = [i.uuid for i in path.involved_items]
    assert actual == expected


def test_PhysicalPath_has_ordered_list_of_involved_links(model: MelodyModel):
    expected = [
        "42ee9e89-d445-45a2-8280-028d4fb1038d",
        "3078ec08-956a-4c61-87ed-0143d1d66715",
    ]

    path = model.pa.all_physical_paths.by_name("card1 - card2 connection")

    actual = [i.uuid for i in path.involved_links]
    assert actual == expected


def test_PhysicalPath_has_exchanges(model: MelodyModel):
    exchange = model.pa.all_component_exchanges.by_name("C 6")
    path = model.pa.all_physical_paths.by_name("card1 - card2 connection")
    assert path.exchanges == [exchange]


def test_PhysicalLink_has_physical_paths(model: MelodyModel):
    link = model.pa.all_physical_links.by_name("Eth Cable 2")
    path = model.pa.all_physical_paths.by_name("card1 - card2 connection")
    assert link.physical_paths == [path]


def test_PhysicalLink_has_exchanges(model: MelodyModel):
    link = model.pa.all_physical_links.by_name("Eth Cable 2")
    exchange = model.pa.all_component_exchanges.by_name("C 3")
    assert link.exchanges == [exchange]


def test_PhysicalLink_setting_ends(model: MelodyModel):
    link = model.pa.all_physical_links.by_name("Eth Cable 2")
    source_pp = model.by_uuid("76d9c301-c0ad-4615-9f02-b804b018decf")
    target_pp = model.by_uuid("53c9fe29-18e2-4642-906e-b7507bf0ff39")

    link.ends = [source_pp, target_pp]

    assert source_pp == link.ends[0]
    assert source_pp == link.source
    assert target_pp == link.ends[1]
    assert target_pp == link.target


def test_PhysicalLink_setting_source_and_target(model: MelodyModel):
    link = model.pa.all_physical_links.by_name("Eth Cable 2")
    source_pp = model.by_uuid("76d9c301-c0ad-4615-9f02-b804b018decf")
    target_pp = model.by_uuid("53c9fe29-18e2-4642-906e-b7507bf0ff39")

    link.source = source_pp
    link.target = target_pp

    assert source_pp == link.ends[0]
    assert source_pp == link.source
    assert target_pp == link.ends[1]
    assert target_pp == link.target



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import errno
import logging
import pathlib
import re
import sys

import pytest

from capellambse import helpers


def test_paths_relative_to_root_are_not_changed():
    actual = helpers.normalize_pure_path("foo", base="/")
    expected = pathlib.PurePosixPath("foo")
    assert actual == expected


def test_paths_relative_to_a_given_base_are_made_relative_to_root():
    actual = helpers.normalize_pure_path("bar", base="/foo")
    expected = pathlib.PurePosixPath("foo/bar")
    assert actual == expected


def test_absolute_paths_are_made_relative_to_root():
    actual = helpers.normalize_pure_path("/foo/bar")
    expected = pathlib.PurePosixPath("foo/bar")
    assert actual == expected


def test_base_is_ignored_for_absolute_paths():
    actual = helpers.normalize_pure_path("/foo", base="/bar")
    expected = pathlib.PurePosixPath("foo")
    assert actual == expected


def test_doubledot_removes_the_preceding_component():
    actual = helpers.normalize_pure_path("/foo/../bar")
    expected = pathlib.PurePosixPath("bar")
    assert actual == expected


def test_leading_doubledots_are_ignored():
    actual = helpers.normalize_pure_path("/../../../foo")
    expected = pathlib.PurePosixPath("foo")
    assert actual == expected


@pytest.mark.parametrize(
    "input,expected_output",
    [("&gt;", ">"), ("&lt;", "<"), ("&quot;", '"'), ("&apos;", "'")],
)
def test_flatten_html_unescapes_special_char(
    input: str, expected_output: str
) -> None:
    assert helpers.flatten_html_string(input) == expected_output


def test_flatten_html_strips_images() -> None:
    input = """<img src="data:<>">"""
    assert helpers.flatten_html_string(input) == ""


@pytest.mark.parametrize(
    "input,expected",
    [
        ("First<br>Second", "First\nSecond"),
        ("<p>First</p>Second", "First\nSecond"),
        (
            "<p>First<img>Second<br>Third</p>Fourth",
            "FirstSecond\nThird\nFourth",
        ),
    ],
)
def test_flatten_html_blocks_and_hardbreaks_to_newline(
    input: str, expected: str
) -> None:
    assert helpers.flatten_html_string(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            "<ul><li>item 1</li><li>item 2</li></ul>",
            "             • item 1\n             • item 2",
        ),
        (
            "\n<ul>\n\t<li>item 1</li>\n\t<li>item 2</li></ul>",
            "             • item 1\n             • item 2",
        ),
    ],
)
def test_flatten_html_formats_unordered_lists(
    input: str, expected: str
) -> None:
    assert helpers.flatten_html_string(input) == expected


def test_process_html_fragments_does_not_process_empty_markup() -> None:
    markup = ""

    def cb(node):
        raise AssertionError("Callback should not be called")

    processed_markup = helpers.process_html_fragments(markup, cb)

    assert str(processed_markup) == ""


def test_process_html_fragments_does_not_process_plain_text() -> None:
    markup = "Test"

    def cb(node):
        raise AssertionError("Callback should not be called")

    processed_markup = helpers.process_html_fragments(markup, cb)

    assert str(processed_markup) == "Test"


def test_process_html_fragments_processes_a_single_root_element() -> None:
    markup = "<div>Test</div>"

    def cb(node):
        if node.text == "Test":
            node.text = "Modified"

    processed_markup = helpers.process_html_fragments(markup, cb)

    assert str(processed_markup) == "<div>Modified</div>"


def test_process_html_fragments_processes_multiple_root_elements() -> None:
    markup = "<div>Test</div><p>Another</p>"

    def cb(node):
        if node.text == "Test":
            node.text = "Modified"
        elif node.text == "Another":
            node.text = "Changed"

    processed_markup = helpers.process_html_fragments(markup, cb)

    assert str(processed_markup) == "<div>Modified</div><p>Changed</p>"


def test_process_html_fragments_keeps_leading_text() -> None:
    markup = "Leading text<div>Test</div>"

    def cb(node):
        if node.text == "Test":
            node.text = "Modified"

    processed_markup = helpers.process_html_fragments(markup, cb)

    assert str(processed_markup) == "Leading text<div>Modified</div>"


def test_process_html_fragments_callback_visits_nodes_in_order() -> None:
    markup = (
        "<div>Outer 1<div>Inner 1</div><div>Inner 2</div></div>"
        "<div>Outer 2<div>Inner 3</div><div>Inner 4</div></div>"
    )
    expected = [
        "Outer 1",
        "Inner 1",
        "Inner 2",
        "Outer 2",
        "Inner 3",
        "Inner 4",
    ]
    actual = []

    def cb(node):
        actual.append(node.text)

    helpers.process_html_fragments(markup, cb)

    assert actual == expected


if sys.platform.startswith("win"):
    import msvcrt

    def test_flock_locks_and_unlocks_again(monkeypatch, tmpdir):
        calls = []

        def mock_locking(_1, mode, _2):
            nonlocal calls
            calls.append(mode)

        monkeypatch.setattr(msvcrt, "locking", mock_locking)
        tmpfile = pathlib.Path(tmpdir, "capellambse.lock")

        with helpers.flock(tmpfile):
            pass

        assert calls == [msvcrt.LK_LOCK, msvcrt.LK_UNLCK]

    def test_flock_retries_after_EDEADLOCK(monkeypatch, tmpdir):
        locks = 0

        def mock_locking(_1, mode, _2):
            nonlocal locks
            if mode == msvcrt.LK_LOCK:
                locks += 1
                if locks < 2:
                    raise OSError(errno.EDEADLOCK, "First non-blocking call")

        monkeypatch.setattr(msvcrt, "locking", mock_locking)
        tmpfile = pathlib.Path(tmpdir, "capellambse.lock")

        with helpers.flock(tmpfile):
            pass

        assert locks == 2

else:
    import fcntl

    def test_flock_blocks_indefinitely(monkeypatch, tmpdir, caplog):
        call_order = []

        def mock_flock(_, flags):
            nonlocal call_order
            if flags & fcntl.LOCK_NB == fcntl.LOCK_NB:
                call_order.append("nonblock")
                raise OSError(errno.EAGAIN, "Non-blocking flock mocks failure")
            else:
                call_order.append("blocking")

        caplog.set_level(logging.DEBUG)
        monkeypatch.setattr(fcntl, "flock", mock_flock)
        tmpfile = pathlib.Path(tmpdir, "capellambse.lock")

        with helpers.flock(tmpfile):
            pass

        logs = [
            i.message
            for i in caplog.records
            if i.name == "capellambse.helpers"
        ]
        assert any(
            re.fullmatch(r"Waiting for lock file /.*/capellambse\.lock", i)
            for i in logs
        )
        assert call_order == ["nonblock", "blocking"]



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Module to populate the diagram cache with native Capella diagrams."""

from __future__ import annotations

import collections
import datetime
import itertools
import json
import logging
import os
import pathlib
import re
import shutil
import typing as t

import lxml
from lxml import etree

import capellambse.helpers
import capellambse.model.diagram
from capellambse import _native
from capellambse.filehandler import local
from capellambse.model import modeltypes

E = lxml.builder.ElementMaker()
BAD_FILENAMES = frozenset(
    {"AUX", "CON", "NUL", "PRN"}
    | {f"COM{i}" for i in range(1, 10)}
    | {f"LPT{i}" for i in range(1, 10)}
)
CROP_MARGIN = 10.0
LOGGER = logging.getLogger(__name__)

SVG_CROP_FLAG_NAME = "CAPELLAMBSE_EXPERIMENTAL_CROP_SVG_DIAGRAM_CACHE_VIEWPORT"
"""Flag environment variable that enables experimental SVG cropping.

If set to 1, the viewBox and width/height of produced SVG images will be
cropped in order to minimize the amount of whitespace around the visible
contents.
"""


VALID_FORMATS = frozenset(
    {
        "bmp",
        "gif",
        "jpg",
        "png",
        "svg",
    }
)
VIEWPOINT_ORDER = (
    "Common",
    "Operational Analysis",
    "System Analysis",
    "Logical Architecture",
    "Physical Architecture",
)


class _NoBoundingBoxFound(Exception):
    pass


class _NoExtentsFound(Exception):
    pass


class ViewBox(t.NamedTuple):
    """Bounding box of an SVG shape."""

    x_min: float
    y_min: float
    width: float
    height: float


class Extents(t.NamedTuple):
    """Extents of an SVG shape."""

    x_min: float
    x_max: float
    y_min: float
    y_max: float


class IndexEntry(t.TypedDict):
    """An entry for the index JSON file."""

    uuid: str
    name: str
    type: modeltypes.DiagramType
    viewpoint: str
    success: bool


def export(
    capella: str,
    model: capellambse.MelodyModel,
    *,
    format: str,
    index: bool,
    force: t.Literal["exe", "docker", None],
    background: bool,
    refresh: bool = False,
) -> None:
    if model.diagram_cache is None:
        raise TypeError("No diagram cache configured for the model")
    if not isinstance(model.diagram_cache, local.LocalFileHandler):
        raise TypeError(
            "Diagram cache updates are only supported for local paths"
        )

    format = format.lower()
    if format not in VALID_FORMATS:
        supported = ", ".join(sorted(VALID_FORMATS))
        raise ValueError(
            f"Invalid image format {format!r}, supported are {supported}"
        )

    diag_cache_dir = pathlib.Path(model.diagram_cache.path)
    try:
        shutil.rmtree(diag_cache_dir)
    except FileNotFoundError:
        pass
    diag_cache_dir.mkdir(parents=True)

    native_args = _find_executor(model, capella, force)
    with _native.native_capella(model, **native_args) as cli:
        if refresh:
            cli(
                *_native.ARGS_CMDLINE,
                "-appid",
                "org.polarsys.capella.refreshRepresentations",
            )

        cli(
            *_native.ARGS_CMDLINE,
            "-appid",
            "org.polarsys.capella.exportRepresentations",
            "-imageFormat",
            format.upper(),
            "-outputfolder",
            "/main_model/output",
            "-forceoutputfoldercreation",
        )

        diagrams = _copy_images(
            model,
            cli.workspace / "main_model" / "output",
            diag_cache_dir,
            format,
            background,
        )
        if index:
            _write_index(model, format, diag_cache_dir, diagrams)


def _find_executor(
    model: capellambse.MelodyModel,
    capella: str,
    force: t.Literal["exe", "docker", None],
) -> dict[str, str]:
    assert model.info.capella_version
    capella = capella.replace("{VERSION}", model.info.capella_version)
    native_args: dict[str, str] = {}
    if force == "docker":
        native_args["docker"] = capella
    elif force == "exe":
        native_args["exe"] = capella
    else:
        if pathlib.Path(capella).is_absolute():
            native_args["exe"] = capella
        elif pathlib.Path(capella).parent == pathlib.Path("."):
            exe = shutil.which(capella)
            if exe:
                native_args["exe"] = exe
            else:
                raise ValueError(f"Not found in PATH: {capella}")
        else:
            native_args["docker"] = capella
    assert "exe" in native_args or "docker" in native_args
    return native_args


def _copy_images(
    model: capellambse.MelodyModel,
    srcdir: pathlib.Path,
    destdir: pathlib.Path,
    extension: str,
    background: bool,
) -> list[IndexEntry]:
    name_counts = collections.defaultdict[str, int](lambda: -1)
    index: list[IndexEntry] = []
    files = {i.name: i for i in srcdir.glob("**/*") if i.is_file()}

    for i in model.diagrams:
        entry: IndexEntry = {
            "uuid": i.uuid,
            "name": i.name,
            "type": i.type,
            "viewpoint": i.viewpoint,
            "success": False,
        }
        index.append(entry)

        name_counts[i.name] = c = name_counts[i.name] + 1
        if c == 0:
            name = _sanitize_filename(i.name + f".{extension}")
        else:
            name = _sanitize_filename(i.name + f"_{c}.{extension}")

        if name not in files:
            continue

        try:
            source = srcdir / files[name]
            destination = destdir / f"{i.uuid}.{extension}"
            if extension == "svg":
                _copy_and_postprocess_svg(source, destination, background)
            else:
                shutil.copyfile(source, destination)

        except Exception:
            LOGGER.exception("Cannot copy diagram %s (%s)", i.name, i.uuid)
        else:
            entry["success"] = True

    return index


def _sanitize_filename(fname: str) -> str:
    r"""Sanitize the filename.

    This function removes all characters that are illegal in file
    names on Windows operating systems, and prefixes reserved names
    with an underscore.

    Note that this also includes the two common directory separators
    (``/`` and ``\``).

    Notes
    -----
    Refer to the `MSDN article about file names`__ for more details.

    __ https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file?redirectedfrom=MSDN#naming-conventions
    """
    fname = fname.rstrip(" .")
    fname = re.sub(
        '[\x00-\x1f<>:"/\\\\|?*]',
        lambda m: "-"[ord(m.group(0)) < 32 :],
        fname,
    )
    if fname.split(".")[0].upper() in BAD_FILENAMES:
        fname = f"_{fname}"
    return fname


def _circle_extents(element: etree._Element) -> Extents:
    """Compute extents for a circle."""
    cx = float(element.get("cx", 0))
    cy = float(element.get("cy", 0))
    r = float(element.get("r", 0))
    return Extents(cx - r, cx + r, cy - r, cy + r)


def _ellipse_extents(element: etree._Element) -> Extents:
    """Compute extents for an ellipse."""
    cx = float(element.get("cx", 0))
    cy = float(element.get("cy", 0))
    rx = float(element.get("rx", 0))
    ry = float(element.get("ry", 0))
    return Extents(cx - rx, cx + rx, cy - ry, cy + ry)


def _image_extents(element: etree._Element) -> Extents:
    """Compute extents for an image."""
    x = float(element.get("x", 0))
    y = float(element.get("y", 0))
    width = float(element.get("width", 0))
    height = float(element.get("height", 0))
    return Extents(x, x + width, y, y + height)


def _line_extents(element: etree._Element) -> Extents:
    """Compute extents for a line."""
    x1 = float(element.get("x1", 0))
    y1 = float(element.get("y1", 0))
    x2 = float(element.get("x2", 0))
    y2 = float(element.get("y2", 0))
    return Extents(min(x1, x2), max(x1, x2), min(y1, y2), max(y1, y2))


def _polyline_extents(element: etree._Element) -> Extents:
    """Compute extents for a polyline or polygon."""
    points = [
        tuple(map(float, p.split(",")))
        for p in element.get("points", "").strip().split()
    ]
    x_coords, y_coords = zip(*points)
    return Extents(min(x_coords), max(x_coords), min(y_coords), max(y_coords))


def _rect_extents(element: etree._Element) -> Extents:
    """Compute extents for a rectangle."""
    width_str = element.get("width", "0")
    height_str = element.get("height", "0")
    try:
        width = float(width_str)
        height = float(height_str)
    except ValueError:
        raise _NoExtentsFound() from None
    x = float(element.get("x", 0))
    y = float(element.get("y", 0))
    width = float(element.get("width", 0))
    height = float(element.get("height", 0))
    return Extents(x, x + width, y, y + height)


def _text_extents(element: etree._Element) -> Extents:
    """Compute extents for text."""
    x = float(element.get("x", 0))
    y = float(element.get("y", 0))
    text = element.text or ""
    font_size_str = element.get("font-size", "12").strip().lower()
    if "px" in font_size_str:
        font_size_str = font_size_str[:-2]
    font_size = float(font_size_str)
    width = len(text) * font_size * 0.6  # Simplistic width estimation
    return Extents(x, x + width, y - font_size, y)


def _use_extents(element: etree._Element) -> Extents:
    """Compute extents for use."""
    x = float(element.get("x", 0))
    y = float(element.get("y", 0))
    width = float(element.get("width", 0))
    height = float(element.get("height", 0))
    return Extents(x, x + width, y, y + height)


EXTENT_FUNCTIONS: dict[str, t.Callable[[etree._Element], Extents]] = {
    "circle": _circle_extents,
    "ellipse": _ellipse_extents,
    "image": _image_extents,
    "line": _line_extents,
    # ``polyline`` and ``polygon`` only differ in whether the shape is
    # implicitly closed (the first and the last point are connected).
    # For the purpose of calculating the extents, this is irrelevant, so
    # we can simply use the same algorithm here.
    "polygon": _polyline_extents,
    "polyline": _polyline_extents,
    "rect": _rect_extents,
    "text": _text_extents,
    "use": _use_extents,
}


def _calculate_svg_viewbox(root: etree._Element) -> ViewBox:
    """Compute bounding box of graphical content for SVG file."""
    x_min = y_min = float("inf")
    x_max = y_max = float("-inf")
    for element in root.iter():
        if (
            isinstance(element, etree._Comment)
            or not hasattr(element, "tag")
            or element.get("transform")
        ):
            continue
        tag = etree.QName(element.tag).localname
        func = EXTENT_FUNCTIONS.get(tag)
        if func is None:
            continue
        try:
            shape_x_min, shape_x_max, shape_y_min, shape_y_max = func(element)
        except (_NoExtentsFound, ValueError):
            continue
        x_min = min(x_min, shape_x_min)
        y_min = min(y_min, shape_y_min)
        x_max = max(x_max, shape_x_max)
        y_max = max(y_max, shape_y_max)

    x_min -= CROP_MARGIN
    y_min -= CROP_MARGIN
    x_max += CROP_MARGIN
    y_max += CROP_MARGIN
    if any(abs(i) == float("inf") for i in (x_min, x_max, y_min, y_max)):
        raise _NoBoundingBoxFound() from None
    return ViewBox(x_min, y_min, x_max - x_min, y_max - y_min)


def _crop_svg_viewbox(src: pathlib.Path, root: etree._Element):
    try:
        min_x, min_y, new_width, new_height = _calculate_svg_viewbox(root)
    except _NoBoundingBoxFound:
        LOGGER.warning("Cannot determine bounding box in file: %s", src)
        return

    old_width = float(root.get("width", 0))
    old_height = float(root.get("height", 0))
    if new_width >= old_width and new_height >= old_height:
        LOGGER.debug(
            (
                "Calculated viewbox for %s is larger than original, ignoring:"
                " (w=%.1f, h=%.1f) > (w=%.1f, h=%.1f)"
            ),
            src,
            new_width,
            new_height,
            old_width,
            old_height,
        )
        return

    old_viewbox_width = float(root.get("viewBox", "0 0 0 0").split()[2])
    old_viewbox_height = float(root.get("viewBox", "0 0 0 0").split()[3])
    if new_width < old_viewbox_width and new_height < old_viewbox_height:
        root.set("viewBox", f"{min_x} {min_y} {new_width} {new_height}")
    elif new_width < old_viewbox_width:
        root.set(
            "viewBox", f"{min_x} {min_y} {new_width} {old_viewbox_height}"
        )
    elif new_height < old_viewbox_height:
        root.set(
            "viewBox", f"{min_x} {min_y} {old_viewbox_width} {new_height}"
        )
    if new_width < old_width:
        root.set("width", str(new_width))
    if new_height < old_height:
        root.set("height", str(new_height))


def _copy_and_postprocess_svg(
    src: pathlib.Path, dest: pathlib.Path, background: bool
) -> None:
    """Copy ``src`` to ``dest`` and post process SVG diagram.

    Post-processing stops propagation of default ``fill`` and ``stroke``
    styling into elements that don't have these stylings. Fixates
    ``font-family`` to ``'Open Sans','Segoe UI',Arial,sans-serif`` and
    deletes ``stroke- miterlimit``.
    """
    tree = etree.parse(src)
    root = tree.getroot()
    if os.getenv(SVG_CROP_FLAG_NAME) == "1":
        _crop_svg_viewbox(src, root)
    if background:
        viewbox = root.get("viewBox", "0 0 0 0").split()
        background_elem = etree.Element(
            "rect",
            x=viewbox[0],
            y=viewbox[1],
            width=viewbox[2],
            height=viewbox[3],
            fill="white",
        )
        root.insert(0, background_elem)
    for elm in tree.iter():
        attrib = elm.attrib
        if "font-family" in attrib:
            attrib["font-family"] = "'Open Sans','Segoe UI',Arial,sans-serif"
        if "stroke-miterlimit" in attrib:
            del attrib["stroke-miterlimit"]
        if elm.tag == "{http://www.w3.org/2000/svg}clipPath":
            attrib.update({"fill": "transparent", "stroke": "transparent"})

    svg = etree.tostring(tree, pretty_print=True)
    dest.write_bytes(svg)


def _write_index(
    model: capellambse.MelodyModel,
    extension: str,
    dest: pathlib.Path,
    index: list[IndexEntry],
) -> None:
    now = datetime.datetime.now().strftime("%A, %Y-%m-%d %H:%M:%S")
    title = f"Capella diagram cache for {model.name!r}"
    html = E.html(
        E.head(
            E.style(
                "a:active, a:hover, a:link, a:visited {text-decoration: none}"
                " .missing {color: red}"
                " .helptext {text-decoration: dashed underline; cursor: help}"
                " .small {font-size: 60%}"
                " .uuid {color: #AAA; font-size: 60%}"
            )
        ),
        body := E.body(E.h1(title), E.p({"class": "small"}, "Created: ", now)),
    )

    def sortkey(entry: IndexEntry):
        try:
            vp_index = VIEWPOINT_ORDER.index(entry["viewpoint"])
        except ValueError:
            vp_index = len(VIEWPOINT_ORDER)
        return (vp_index, entry["viewpoint"], entry["name"])

    diagrams = sorted(index, key=sortkey)
    for vp, diags in itertools.groupby(diagrams, key=lambda i: i["viewpoint"]):
        body.append(E.h2(vp))
        body.append(ol := E.ol())
        for diagram in diags:
            uuid = E.span(diagram["uuid"], {"class": "uuid"})
            tlabel = E.span(
                {"title": diagram["type"].value, "class": "helptext"},
                f"[{diagram['type'].name}]",
            )
            if diagram["success"]:
                href = f"{diagram['uuid']}.{extension}"
                label = E.a({"href": href}, diagram["name"])
            else:
                label = E.span({"class": "missing"}, diagram["name"])
            ol.append(E.li(tlabel, " ", label, " ", uuid))

    dest.joinpath("index.json").write_text(json.dumps(index, cls=IndexEncoder))
    dest.joinpath("index.html").write_bytes(etree.tostring(html))


class IndexEncoder(json.JSONEncoder):
    """A JSON encoder for the index file."""

    def default(self, o):
        if isinstance(o, modeltypes.DiagramType):
            return o.name
        return super().default(o)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = ["UnsupportedPluginError", "UnsupportedPluginVersionError"]

import collections.abc as cabc
import dataclasses
import re
import typing as t


class UnsupportedPluginError(ValueError):
    """Raised when a plugin is unknown."""


class UnsupportedPluginVersionError(ValueError):
    """Raised when the plugin's version is unsupported."""

    def __str__(self) -> str:
        plugin, ns_plugin = self.args
        assert isinstance(plugin, Plugin) and isinstance(ns_plugin, Plugin)
        return (
            f"Plugin '{plugin.name}' with unsupported version encountered: "
            f"{plugin.version!r} does not match {ns_plugin.version!r}."
        )


@dataclasses.dataclass(frozen=True)
class Plugin:
    """Capella xml-element/plugin with info about name and version."""

    name: str
    version: str | tuple[str, str] | None = None
    viewpoint: str | None = None

    version_precision: int = 1
    """Number of significant parts in the version number for namespaces.

    When generating a versioned namespace URL from a Plugin, only this
    many digits will be taken from the activated viewpoint. This allows
    reusing the same namespace across subsequent minor plugin releases.

    Example: A version of "1.2.3" with precision set to 2 will result in
    the namespace version "1.2.0" being used.

    Note that the used version number will always be padded with zeroes
    to as many parts as there are in the original version number.
    """

    def __post_init__(self) -> None:
        if self.version is not None and self.viewpoint is None:
            raise TypeError("Versioned plugins require a viewpoint")

    @property
    def min_version(self) -> str | None:
        if isinstance(self.version, str):
            return self.version
        if isinstance(self.version, tuple):
            return self.version[0]
        return None

    @property
    def max_version(self) -> str | None:
        if isinstance(self.version, str):
            return self.version
        if isinstance(self.version, tuple):
            return self.version[1]
        return None

    def __le__(self, other: Plugin) -> bool:
        if self.version is None:
            raise AttributeError(f"Plugin '{self.name}' has no version")

        if isinstance(other.version, tuple):
            assert all(isinstance(o, str) for o in other.version)
            vmin, vmax = tuple(map(_tofloat, other.version))
            if isinstance(self.version, tuple):
                vsmin, vsmax = tuple(map(_tofloat, self.version))
                return vsmin <= vmin <= vmax <= vsmax
            else:
                return vmin <= _tofloat(self.version) <= vmax

        assert other.version is not None
        oversion = _tofloat(other.version)
        if isinstance(self.version, tuple):
            vsmin, vsmax = tuple(map(_tofloat, self.version))
            return vsmin <= oversion <= vsmax

        return oversion <= _tofloat(self.version)

    def __str__(self) -> str:
        if suffix := self.min_version:
            return self.name + suffix
        return self.name

    def matches_version(self, value: str | None) -> bool:
        """Check whether a version number is supported by this plugin."""
        if value is None:
            return self.version is None
        if self.version is None:
            return True
        if isinstance(self.version, str):
            return self.version == value

        for mymin, mymax, their in zip(
            self.version[0].split("."),
            self.version[1].split("."),
            value.split("."),
        ):
            if not int(mymin) <= int(their) <= int(mymax):
                return False
        return True


def _tofloat(other: str) -> float:
    """Change 1.x.y...str into 1.x float."""
    version = other.split(".")
    if len(version) > 1:
        version = version[:2]

    return float(".".join(version))


def get_namespace_prefix(url: str) -> str:
    """Map a namespace URL to its symbolic name (the "prefix").

    Parameters
    ----------
    url
        A namespace URL.

    Raises
    ------
    UnsupportedPluginError
        If the given URL does not match any known plugin.
    UnsupportedPluginVersionError
        If the URL matches a known plugin, but the version indicated by
        the URL is outside of the supported range.

    Returns
    -------
    str
        The symbolic name of the namespace matching the URI.
    """
    match = re.match(r"(.*/)((?:\d+\.)*\d+)$", url)
    if match:
        plugin_name, version = match.groups()
    else:
        plugin_name, version = url, None

    matched_plugins = [
        (nskey, nsplugin)
        for nskey, nsplugin in NAMESPACES_PLUGINS.items()
        if nsplugin.name in (plugin_name, url)
    ]
    if not matched_plugins:
        raise UnsupportedPluginError(url)
    if len(matched_plugins) != 1:
        raise RuntimeError(f"Ambiguous namespace {url!r}: {matched_plugins}")
    prefix, plugin = matched_plugins[0]

    if not plugin.matches_version(version):
        raise UnsupportedPluginVersionError(plugin_name, version)

    return prefix


NAMESPACES_PLUGINS: t.Final[cabc.Mapping[str, Plugin]] = {
    "CapellaRequirements": Plugin(
        "http://www.polarsys.org/capella/requirements",
        None,
        "org.polarsys.capella.vp.requirements",
    ),
    "Requirements": Plugin(
        "http://www.polarsys.org/kitalpha/requirements",
        None,
        "org.polarsys.kitalpha.vp.requirements",
    ),
    "concern": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/concern/1.1.0",
    ),
    "description": Plugin("http://www.eclipse.org/sirius/description/1.1.0"),
    "description_1": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/1.1.0"
    ),
    "description_2": Plugin(
        "http://www.eclipse.org/sirius/table/description/1.1.0"
    ),
    "description_3": Plugin(
        "http://www.eclipse.org/sirius/diagram/sequence/description/2.0.0",
    ),
    "diagram": Plugin("http://www.eclipse.org/sirius/diagram/1.1.0"),
    "diagramstyler": Plugin("http://thalesgroup.com/mde/melody/ordering"),
    "ecore": Plugin("http://www.eclipse.org/emf/2002/Ecore"),
    "filter": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/filter/1.1.0",
    ),
    "libraries": Plugin(
        "http://www.polarsys.org/capella/common/libraries/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "metadata": Plugin("http://www.polarsys.org/kitalpha/ad/metadata/1.0.0"),
    "notation": Plugin("http://www.eclipse.org/gmf/runtime/1.0.2/notation"),
    "org.polarsys.capella.core.data.capellacommon": Plugin(
        "http://www.polarsys.org/capella/core/common/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.capellacore": Plugin(
        "http://www.polarsys.org/capella/core/core/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.capellamodeller": Plugin(
        "http://www.polarsys.org/capella/core/modeller/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.cs": Plugin(
        "http://www.polarsys.org/capella/core/cs/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.ctx": Plugin(
        "http://www.polarsys.org/capella/core/ctx/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.epbs": Plugin(
        "http://www.polarsys.org/capella/core/epbs/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.fa": Plugin(
        "http://www.polarsys.org/capella/core/fa/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.information": Plugin(
        "http://www.polarsys.org/capella/core/information/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.information.datatype": Plugin(
        "http://www.polarsys.org/capella/core/information/datatype/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.information.datavalue": Plugin(
        "http://www.polarsys.org/capella/core/information/datavalue/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.interaction": Plugin(
        "http://www.polarsys.org/capella/core/interaction/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.la": Plugin(
        "http://www.polarsys.org/capella/core/la/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.oa": Plugin(
        "http://www.polarsys.org/capella/core/oa/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "org.polarsys.capella.core.data.pa": Plugin(
        "http://www.polarsys.org/capella/core/pa/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "re": Plugin(
        "http://www.polarsys.org/capella/common/re/",
        ("5.0.0", "6.0.0"),
        "org.polarsys.capella.core.viewpoint",
    ),
    "sequence": Plugin("http://www.eclipse.org/sirius/diagram/sequence/2.0.0"),
    "style": Plugin(
        "http://www.eclipse.org/sirius/diagram/description/style/1.1.0"
    ),
    "table": Plugin("http://www.eclipse.org/sirius/table/1.1.0"),
    "viewpoint": Plugin("http://www.eclipse.org/sirius/1.1.0"),
    "xmi": Plugin("http://www.omg.org/XMI"),
    "xsi": Plugin("http://www.w3.org/2001/XMLSchema-instance"),
}
NAMESPACES: t.Final[cabc.Mapping[str, str]] = {
    nskey: str(plugin) for nskey, plugin in NAMESPACES_PLUGINS.items()
}



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = [
    "InvalidModificationError",
    "NonUniqueMemberError",
    "Accessor",
    "DeprecatedAccessor",
    "WritableAccessor",
    "DirectProxyAccessor",
    "DeepProxyAccessor",
    "LinkAccessor",
    "AttrProxyAccessor",
    "PhysicalLinkEndsAccessor",
    "IndexAccessor",
    "AlternateAccessor",
    "ParentAccessor",
    "CustomAccessor",
    "AttributeMatcherAccessor",
    "SpecificationAccessor",
    "ReferenceSearchingAccessor",
    "ElementListCouplingMixin",
    "RoleTagAccessor",
    "TypecastAccessor",
    "PhysicalAccessor",
]

import abc
import collections.abc as cabc
import contextlib
import itertools
import logging
import operator
import sys
import typing as t
import warnings

import markupsafe
import typing_extensions as te
from lxml import etree

import capellambse
from capellambse import helpers

from . import XTYPE_HANDLERS, S, T, U, build_xtype, element

_NOT_SPECIFIED = object()
"Used to detect unspecified optional arguments"

LOGGER = logging.getLogger(__name__)


class InvalidModificationError(RuntimeError):
    """Raised when a modification would result in an invalid model."""


class NonUniqueMemberError(ValueError):
    """Raised when a duplicate member is inserted into a list."""

    parent = property(lambda self: self.args[0])
    attr = property(lambda self: self.args[1])
    target = property(lambda self: self.args[2])

    def __str__(self) -> str:
        if len(self.args) != 3:
            return super().__str__()

        return (
            f"Cannot insert: {self.attr!r} of {self.parent._short_repr_()}"
            f" already contains a reference to {self.target._short_repr_()}"
        )


class _NewObject:
    """A marker class that indicates that a new object should be created.

    This object can be assigned to an attribute of a model object, and
    will be replaced with a new object of the correct type by the
    attribute's accessor.

    In the future, this class will be replaced with a function that
    simply returns a new object of the correct type.
    """

    def __init__(self, /, *type_hint: str, **kw: t.Any) -> None:
        self._type_hint = type_hint
        self._kw = kw

    def __repr__(self) -> str:
        kw = ", ".join(f"{k}={v!r}" for k, v in self._kw.items())
        return f"<new object {self._type_hint!r} ({kw})>"


class Accessor(t.Generic[T], metaclass=abc.ABCMeta):
    """Super class for all Accessor types."""

    __objclass__: type[t.Any]
    __name__: str

    def __init__(self) -> None:
        super().__init__()
        self.__doc__ = (
            f"A {type(self).__name__} that was not properly configured."
            " Ensure that ``__set_name__`` gets called after construction."
        )

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> te.Self: ...

    @t.overload
    def __get__(
        self, obj: element.ModelObject, objtype: type[t.Any] | None = ...
    ) -> T | element.ElementList[T]: ...

    @abc.abstractmethod
    def __get__(
        self,
        obj: element.ModelObject | None,
        objtype: type[t.Any] | None = None,
    ) -> te.Self | T | element.ElementList[T]:
        pass

    def __set__(self, obj: element.ModelObject, value: t.Any) -> None:
        raise TypeError("Cannot set this type of attribute")

    def __delete__(self, obj: element.ModelObject) -> None:
        raise TypeError("Cannot delete this type of attribute")

    def __set_name__(self, owner: type[t.Any], name: str) -> None:
        self.__objclass__ = owner
        self.__name__ = name
        friendly_name = name.replace("_", " ")
        self.__doc__ = f"The {friendly_name} of this {owner.__name__}."

    def __repr__(self) -> str:
        return f"<{self._qualname!r} {type(self).__name__}>"

    @property
    def _qualname(self) -> str:
        """Generate the qualified name of this descriptor."""
        if not hasattr(self, "__objclass__"):
            return f"(unknown {type(self).__name__} - call __set_name__)"
        return f"{self.__objclass__.__name__}.{self.__name__}"


class DeprecatedAccessor(Accessor[T]):
    """Provides a deprecated alias to another attribute."""

    __slots__ = ("alternative",)

    def __init__(self, alternative: str, /) -> None:
        super().__init__()
        self.alternative = alternative

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> te.Self: ...

    @t.overload
    def __get__(
        self,
        obj: element.ModelObject,
        objtype: type[t.Any] | None = ...,
    ) -> T | element.ElementList[T]: ...

    def __get__(
        self,
        obj: element.ModelObject | None,
        objtype: type[t.Any] | None = None,
    ) -> te.Self | T | element.ElementList[T]:
        if obj is None:
            return self

        self.__warn()
        return getattr(obj, self.alternative)

    def __set__(self, obj: element.ModelObject, value: t.Any) -> None:
        self.__warn()
        setattr(obj, self.alternative, value)

    def __delete__(self, obj: element.ModelObject) -> None:
        self.__warn()
        delattr(obj, self.alternative)

    def __warn(self) -> None:
        msg = f"{self._qualname} is deprecated, use {self.alternative} instead"
        warnings.warn(msg, FutureWarning, stacklevel=3)


class WritableAccessor(Accessor[T], metaclass=abc.ABCMeta):
    """An Accessor that also provides write support on lists it returns."""

    aslist: type[ElementListCouplingMixin] | None
    class_: type[T]
    list_extra_args: cabc.Mapping[str, t.Any]
    single_attr: str | None

    def __init__(
        self,
        *args: t.Any,
        aslist: type[element.ElementList] | None,
        single_attr: str | None = None,
        **kw: t.Any,
    ) -> None:
        super().__init__(*args, **kw)
        self.single_attr = single_attr
        if aslist is not None:
            self.aslist = type(
                "Coupled" + aslist.__name__,
                (ElementListCouplingMixin, aslist),
                {"_accessor": self},
            )
            self.aslist.__module__ = __name__
        else:
            self.aslist = None

    def __set__(
        self,
        obj: element.ModelObject,
        value: T | _NewObject | cabc.Iterable[T | _NewObject],
    ) -> None:
        raise TypeError("Cannot set this type of attribute")

    def create(
        self,
        elmlist: ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> T:
        """Create and return a new element of type ``elmclass``.

        Parameters
        ----------
        elmlist
            The (coupled)
            :py:class:`~capellambse.model.common.element.ElementList` to
            insert the new object into.
        type_hints
            Hints for finding the correct type of element to create. Can
            either be a full or shortened ``xsi:type`` string, or an
            abbreviation defined by the specific Accessor instance.
        kw
            Initialize the properties of the new object. Depending on
            the object's type, some attributes may be required.
        """
        raise TypeError("Cannot create objects")

    def create_singleattr(
        self, elmlist: ElementListCouplingMixin, arg: t.Any, /
    ) -> T:
        """Create an element that only has a single attribute of interest."""
        if self.single_attr is None:
            raise TypeError(
                "Cannot create object from string, a dictionary is required"
            )
        return self.create(elmlist, **{self.single_attr: arg})

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject | _NewObject,
    ) -> None:
        """Insert the ``value`` object into the model.

        The object must be inserted at an appropriate place, so that, if
        ``elmlist`` were to be created afresh, ``value`` would show up
        at index ``index``.
        """
        raise NotImplementedError("Objects cannot be inserted into this list")

    def delete(
        self,
        elmlist: ElementListCouplingMixin,
        obj: element.ModelObject,
    ) -> None:
        """Delete the ``obj`` from the model."""
        raise NotImplementedError("Objects in this list cannot be deleted")

    def _create(
        self,
        parent: element.ModelObject,
        xmltag: str | None,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> T:
        if type_hints:
            elmclass, kw["xtype"] = self._match_xtype(*type_hints)
        else:
            elmclass, kw["xtype"] = self._guess_xtype()
        assert elmclass is not None

        want_id: str | None = None
        if "uuid" in kw:
            want_id = kw.pop("uuid")

        pelem = parent._element
        with parent._model._loader.new_uuid(pelem, want=want_id) as obj_id:
            obj = elmclass(parent._model, pelem, xmltag, uuid=obj_id, **kw)
        return obj

    def _make_list(self, parent_obj, elements):
        assert hasattr(self, "class_")
        assert hasattr(self, "list_extra_args")
        if self.aslist is None:
            return no_list(self, parent_obj._model, elements, self.class_)
        return self.aslist(
            parent_obj._model,
            elements,
            self.class_,
            parent=parent_obj,
            **self.list_extra_args,
        )

    def _match_xtype(
        self,
        type_1: str | None,
        type_2: str | object = _NOT_SPECIFIED,
        /,
    ) -> tuple[type[T], str]:
        r"""Find the right class for the given ``xsi:type``\ (s)."""

        def match_xt(xtp: S, itr: cabc.Iterable[S]) -> S:
            matches: list[S] = []
            for i in itr:
                if (
                    xtp is i is None
                    or i is not None
                    and xtp == i
                    or xtp == i.split(":")[-1]  # type: ignore[union-attr]
                ):
                    matches.append(i)
            if not matches:
                raise ValueError(f"Invalid or unknown xsi:type {xtp!r}")
            if len(matches) > 1:
                raise ValueError(
                    f"Ambiguous xsi:type {xtp!r}, please qualify: "
                    + ", ".join(repr(i) for i in matches)
                )
            return matches[0]

        if not isinstance(type_1, str):
            raise TypeError(
                f"Expected str as first type, got {type(type_1).__name__!r}"
            )
        candidate_classes: dict[str, type[T]]
        if type_2 is _NOT_SPECIFIED:
            candidate_classes = dict(
                itertools.chain.from_iterable(
                    i.items() for i in XTYPE_HANDLERS.values()
                )
            )
            objtype = type_1
        elif not isinstance(type_2, str):
            raise TypeError(
                f"Expected a str objtype, not {type(type_2).__name__}"
            )
        else:
            candidate_classes = XTYPE_HANDLERS[
                match_xt(type_1, XTYPE_HANDLERS)
            ]
            objtype = type_2

        objtype = match_xt(objtype, candidate_classes)
        return candidate_classes[objtype], objtype

    def _guess_xtype(self) -> tuple[type[T], str]:
        try:
            super_guess = super()._guess_xtype  # type: ignore[misc]
        except AttributeError:
            pass
        else:
            return super_guess()
        raise TypeError(f"{self._qualname} requires a type hint")

    def purge_references(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> contextlib.AbstractContextManager[None]:
        """Purge references to the given object from the model.

        This method is called while deleting physical objects, in order
        to get rid of references to that object (and its descendants).
        Reference purging is done in two steps, which is why this method
        returns a context manager.

        The first step, executed by the ``__enter__`` method, collects
        references to the target and ensures that deleting them would
        result in a valid model. If any validity constraints would be
        violated, an exception is raised to indicate as such, and the
        whole operation is aborted. This is also when the relevant
        "capellambse.delete" audit events are fired for each reference.

        Once all ``__enter__`` methods have been called, the target
        object is deleted from the model. Then all ``__exit__`` methods
        are called, which triggers the actual deletion of all previously
        discovered references.

        As per the context manager protocol, ``__exit__`` will always be
        called after ``__enter__``, even if the operation is to be
        aborted. The ``__exit__`` method must therefore inspect whether
        an exception was passed in or not in order to know whether the
        operation succeeded.

        In order to not confuse other context managers and keep the
        model consistent, ``__exit__`` must not raise any further
        exceptions. Exceptions should instead be logged to stderr, for
        example by using the
        :external:py:meth:`logging.Logger.exception` facility.

        The ``purge_references`` method will only be called for Accessor
        instances that actually contain a reference.

        Parameters
        ----------
        obj
            The model object to purge references from.
        target
            The object that is to be deleted; references to this object
            will be purged.

        Returns
        -------
        contextlib.AbstractContextManager
            A context manager that deals with purging references in a
            transactional manner.

        Raises
        ------
        InvalidModificationError
            Raised by the returned context manager's ``__enter__``
            method if the attempted modification would result in an
            invalid model. Note that it is generally preferred to allow
            the operation and take the necessary steps to keep the model
            consistent, if possible. This can be achieved for example by
            deleting dependent objects along with the original deletion
            target.
        Exception
            Any exception may be raised before ``__enter__`` returns in
            order to abort the transaction and prevent the ``obj`` from
            being deleted. No exceptions must be raised by ``__exit__``.

        Examples
        --------
        A simple implementation for purging a single object reference
        could look like this:

        .. code-block:: python

           @contextlib.contextmanager
           def purge_references(self, obj, target):
               assert self.__get__(obj, type(obj)) == target
               sys.audit("capellambse.delete", obj, self.__name__, None)

               yield

               try:
                   self.__delete__(obj)
               except Exception:
                   LOGGER.exception("Could not purge a dangling reference")
        """
        raise NotImplementedError(
            f"{type(self).__name__} does not support purging references"
        )


class PhysicalAccessor(Accessor[T]):
    """Helper super class for accessors that work with real elements."""

    __slots__ = (
        "aslist",
        "class_",
        "list_extra_args",
        "xtypes",
    )

    aslist: type[element.ElementList] | None
    class_: type[T]
    list_extra_args: cabc.Mapping[str, t.Any]
    xtypes: cabc.Set[str]

    def __init__(
        self,
        class_: type[T],
        xtypes: (
            str
            | type[element.ModelObject]
            | cabc.Iterable[str | type[element.ModelObject]]
            | None
        ) = None,
        *,
        aslist: type[element.ElementList[T]] | None = None,
        list_extra_args: cabc.Mapping[str, t.Any] | None = None,
    ) -> None:
        super().__init__()
        if xtypes is None:
            self.xtypes = (
                {build_xtype(class_)}
                if class_ is not element.GenericElement
                else set()
            )
        elif isinstance(xtypes, type):
            assert issubclass(xtypes, element.GenericElement)
            self.xtypes = {build_xtype(xtypes)}
        elif isinstance(xtypes, str):
            self.xtypes = {xtypes}
        else:
            self.xtypes = {
                i if isinstance(i, str) else build_xtype(i) for i in xtypes
            }

        self.aslist = aslist
        self.class_ = class_
        self.list_extra_args = list_extra_args or {}

    def _guess_xtype(self) -> tuple[type[T], str]:
        """Try to guess the type of element that should be created."""
        if self.class_ is element.GenericElement or self.class_ is None:
            raise ValueError("Multiple object types that can be created")
        if not self.xtypes:
            raise ValueError("No matching `xsi:type` found")
        if len(self.xtypes) > 1:
            raise ValueError("Multiple matching `xsi:type`s")
        return self.class_, next(iter(self.xtypes))

    def _make_list(self, parent_obj, elements):
        if self.aslist is None:
            return no_list(self, parent_obj._model, elements, self.class_)
        return self.aslist(
            parent_obj._model,
            elements,
            self.class_,
            **self.list_extra_args,
        )


class DirectProxyAccessor(WritableAccessor[T], PhysicalAccessor[T]):
    """Creates proxy objects on the fly."""

    __slots__ = ("follow_abstract", "rootelem")

    aslist: type[ElementListCouplingMixin] | None
    class_: type[T]
    single_attr: str | None

    def __init__(
        self,
        class_: type[T],
        xtypes: str | type[T] | cabc.Iterable[str | type[T]] | None = None,
        *,
        aslist: type[element.ElementList] | None = None,
        follow_abstract: bool = False,
        list_extra_args: dict[str, t.Any] | None = None,
        rootelem: (
            str
            | type[element.GenericElement]
            | cabc.Sequence[str | type[element.GenericElement]]
            | None
        ) = None,
        single_attr: str | None = None,
    ):
        r"""Create a DirectProxyAccessor.

        Parameters
        ----------
        class_
            The proxy class.
        xtypes
            The ``xsi:type``\ (s) of the child element(s). If None, then
            the constructed proxy will be passed the original element
            instead of a child.
        aslist
            If None, only a single element must match, which will be
            returned directly. If not None, must be a subclass of
            :class:`~capellambse.model.common.element.ElementList`,
            which will be used to return a list of all matched objects.
        follow_abstract
            Follow the link in the ``abstractType`` XML attribute of
            each list member and instantiate that object instead. The
            default is to instantiate the child elements directly.
        list_extra_args
            Extra arguments to pass to the
            :class:`~capellambse.model.common.element.ElementList`
            constructor.
        rootelem
            A class or ``xsi:type`` (or list thereof) that defines the
            path from the current object's XML element to the search
            root. If None, the current element will be used directly.
        single_attr
            If objects can be created with only a single attribute
            specified, this argument is the name of that attribute. This
            :meth:`~capellambse.model.common.accessors\
            .WritableAccessor.create_singleattr`.
        """
        super().__init__(
            class_,
            xtypes,
            aslist=aslist,
            list_extra_args=list_extra_args,
            single_attr=single_attr,
        )
        self.follow_abstract: bool = follow_abstract
        if rootelem is None:
            self.rootelem: cabc.Sequence[str] = []
        elif isinstance(rootelem, str):
            self.rootelem = rootelem.split("/")
        elif isinstance(rootelem, type) and issubclass(
            rootelem, element.GenericElement
        ):
            self.rootelem = [build_xtype(rootelem)]
        else:
            self.rootelem = [
                i if isinstance(i, str) else build_xtype(i) for i in rootelem
            ]

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        elems = [
            self._resolve(obj, e)
            for e in self._getsubelems(obj)
            if e.get("id") is not None
        ]
        rv = self._make_list(obj, elems)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def __set__(
        self,
        obj: element.ModelObject,
        value: str | T | _NewObject | cabc.Iterable[str | T | _NewObject],
    ) -> None:
        if getattr(obj, "_constructed", True):
            sys.audit("capellambse.setattr", obj, self.__name__, value)

        if self.aslist:
            if isinstance(value, str) or not isinstance(value, cabc.Iterable):
                raise TypeError("Can only set list attribute to an iterable")
            list = self.__get__(obj)
            for v in list:
                self.delete(list, v)
            for i, v in enumerate(value):
                if isinstance(v, str):
                    self.create_singleattr(list, v)
                else:
                    self.insert(list, i, v)
        else:
            if isinstance(value, cabc.Iterable) and not isinstance(value, str):
                raise TypeError("Cannot set non-list attribute to an iterable")
            if not isinstance(value, _NewObject):
                raise NotImplementedError(
                    "Moving model objects is not supported yet"
                )
            if self.__get__(obj) is not None:
                raise NotImplementedError(
                    "Replacing model objects is not supported yet"
                )
            self._create(obj, None, *value._type_hint, **value._kw)

    def __delete__(self, obj: element.ModelObject) -> None:
        if self.rootelem:
            raise TypeError("Cannot delete due to 'rootelem' being set")
        if self.follow_abstract:
            raise TypeError("Cannot delete when following abstract types")

        if getattr(obj, "_constructed", True):
            sys.audit("capellambse.delete", obj, self.__name__, None)

        if self.aslist is not None:
            self._delete(obj._model, list(self._getsubelems(obj)))
        else:
            raise TypeError(f"Cannot delete {self._qualname}")

    def _delete(
        self, model: capellambse.MelodyModel, elements: list[etree._Element]
    ) -> None:
        all_elements = (
            list(
                itertools.chain.from_iterable(
                    model._loader.iterdescendants_xt(i) for i in elements
                )
            )
            + elements
        )
        with contextlib.ExitStack() as stack:
            for elm in all_elements:
                if elm.get("id") is None:
                    continue

                obj = element.GenericElement.from_model(model, elm)
                for ref, attr, _ in model.find_references(obj):
                    acc = getattr(type(ref), attr)
                    if acc is self or not isinstance(acc, WritableAccessor):
                        continue
                    stack.enter_context(acc.purge_references(ref, obj))

            for elm in elements:
                parent = elm.getparent()
                assert parent is not None
                model._loader.idcache_remove(elm)
                parent.remove(elm)

    def _resolve(
        self, obj: element.ModelObject, elem: etree._Element
    ) -> etree._Element:
        if self.follow_abstract:
            if abstype := elem.get("abstractType"):
                elem = obj._model._loader[abstype]
            else:
                raise RuntimeError("Broken XML: No abstractType defined?")
        return elem

    def _getsubelems(
        self, obj: element.ModelObject
    ) -> cabc.Iterator[etree._Element]:
        return itertools.chain.from_iterable(
            obj._model._loader.iterchildren_xt(i, *iter(self.xtypes))
            for i in self._findroots(obj)
        )

    def _findroots(self, obj: element.ModelObject) -> list[etree._Element]:
        roots = [obj._element]
        for xtype in self.rootelem:
            roots = list(
                itertools.chain.from_iterable(
                    obj._model._loader.iterchildren_xt(i, xtype) for i in roots
                )
            )
        return roots

    def create(
        self,
        elmlist: ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> T:
        if self.rootelem:
            raise TypeError("Cannot create objects here")

        return self._create(elmlist._parent, None, *type_hints, **kw)

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject | _NewObject,
    ) -> None:
        if isinstance(value, _NewObject):
            raise NotImplementedError(
                "Creating new objects in lists with new_object() is not"
                " supported yet"
            )

        if value._model is not elmlist._model:
            raise ValueError("Cannot move elements between models")
        try:
            indexof = elmlist._parent._element.index
            if index > 0:
                parent_index = indexof(elmlist._elements[index - 1]) + 1
            elif index < -1:
                parent_index = indexof(elmlist._elements[index + 1]) - 1
            else:
                parent_index = index
        except ValueError:
            parent_index = len(elmlist._parent._element)
        elmlist._parent._element.insert(parent_index, value._element)
        elmlist._model._loader.idcache_index(value._element)

    def delete(
        self,
        elmlist: ElementListCouplingMixin,
        obj: element.ModelObject,
    ) -> None:
        assert obj._model is elmlist._model
        self._delete(obj._model, [obj._element])

    @contextlib.contextmanager
    def purge_references(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> cabc.Iterator[None]:
        yield


class DeepProxyAccessor(DirectProxyAccessor[T]):
    """A DirectProxyAccessor that searches recursively through the tree."""

    __slots__ = ()

    def _getsubelems(
        self, obj: element.ModelObject
    ) -> cabc.Iterator[etree._Element]:
        return itertools.chain.from_iterable(
            obj._model._loader.iterdescendants_xt(i, *iter(self.xtypes))
            for i in self._findroots(obj)
        )


class LinkAccessor(WritableAccessor[T], PhysicalAccessor[T]):
    """Accesses elements through reference elements."""

    __slots__ = ("backattr", "tag", "unique")

    aslist: type[ElementListCouplingMixin] | None
    backattr: str | None
    class_: type[T]
    tag: str | None

    def __init__(
        self,
        tag: str | None,
        xtype: str | type[element.GenericElement],
        /,
        *,
        aslist: type[element.ElementList] | None = None,
        attr: str,
        backattr: str | None = None,
        unique: bool = True,
    ) -> None:
        """Create a LinkAccessor.

        Parameters
        ----------
        tag
            The XML tag that the reference elements will have.
        xtype
            The ``xsi:type`` that the reference elements will have. This
            has no influence on the elements that are referenced.
        attr
            The attribute on the reference element that contains the
            actual link.
        backattr
            An optional attribute on the reference element to store a
            reference back to the owner (parent) object.
        aslist
            Optionally specify a different subclass of
            :class:`~capellambse.model.common.element.ElementList`.
        unique
            Enforce that each element may only appear once in the list.
            If a duplicate is attempted to be added, an exception will
            be raised. Note that this does not have an effect on lists
            already existing within the loaded model.
        """
        if not tag:
            warnings.warn(
                "Unspecified XML tag is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
        elif not isinstance(tag, str):
            raise TypeError(f"tag must be a str, not {type(tag).__name__}")
        super().__init__(element.GenericElement, xtype, aslist=aslist)
        if len(self.xtypes) != 1:
            raise TypeError(f"One xtype is required, got {len(self.xtypes)}")
        self.follow = attr
        self.backattr = backattr
        self.tag = tag
        self.unique = unique

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        elems: list[etree._Element] = []
        seen: set[etree._Element] = set()
        for i in self.__find_refs(obj):
            e = self.__follow_ref(obj, i)
            if e is None:
                continue
            if e not in seen:
                elems.append(e)
                seen.add(e)
        rv = self._make_list(obj, elems)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def __set__(self, obj: element.ModelObject, value: t.Any) -> None:
        if obj._constructed:
            sys.audit("capellambse.setattr", obj, self.__name__, value)
        if self.aslist is None:
            if isinstance(value, cabc.Iterable) and not isinstance(value, str):
                raise TypeError(f"{self._qualname} expects a single element")
            value = (value,)
        elif isinstance(value, str) or not isinstance(value, cabc.Iterable):
            raise TypeError(f"{self._qualname} expects an iterable")
        if self.tag is None:
            raise NotImplementedError("Cannot set: XML tag not set")

        self.__delete__(obj)
        for v in value:
            self.__create_link(obj, v)

    def __delete__(self, obj: element.ModelObject) -> None:
        if getattr(obj, "_constructed", True):
            sys.audit("capellambse.delete", obj, self.__name__, None)

        refobjs = list(self.__find_refs(obj))
        for i in refobjs:
            obj._model._loader.idcache_remove(i)
            obj._element.remove(i)

    def __follow_ref(
        self, obj: element.ModelObject, refelm: etree._Element
    ) -> etree._Element | None:
        link = refelm.get(self.follow)
        if not link:
            return None
        return obj._model._loader.follow_link(obj._element, link)

    def __find_refs(
        self, obj: element.ModelObject
    ) -> cabc.Iterator[etree._Element]:
        for refelm in obj._element.iterchildren(tag=self.tag):
            if helpers.xtype_of(refelm) in self.xtypes:
                yield refelm

    def __backref(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> etree._Element | None:
        for i in self.__find_refs(obj):
            if self.__follow_ref(obj, i) == target._element:
                return i
        return None

    def __create_link(
        self,
        parent: element.ModelObject,
        target: element.ModelObject,
        *,
        before: element.ModelObject | None = None,
    ) -> etree._Element:
        assert self.tag is not None
        loader = parent._model._loader

        if self.unique:
            for i in self.__find_refs(parent):
                if self.__follow_ref(parent, i) is target._element:
                    raise NonUniqueMemberError(parent, self.__name__, target)

        with loader.new_uuid(parent._element) as obj_id:
            (xtype,) = self.xtypes
            link = loader.create_link(parent._element, target._element)
            refobj = parent._element.makeelement(
                self.tag,
                {helpers.ATT_XT: xtype, "id": obj_id, self.follow: link},
            )
            if self.backattr and (parent_id := parent._element.get("id")):
                refobj.set(self.backattr, f"#{parent_id}")
            if before is None:
                parent._element.append(refobj)
            else:
                before_elm = self.__backref(parent, before)
                assert before_elm is not None
                assert before_elm in parent._element
                before_elm.addprevious(refobj)
            loader.idcache_index(refobj)
        return refobj

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject | _NewObject,
    ) -> None:
        if self.aslist is None:
            raise TypeError("Cannot insert: This is not a list (bug?)")
        if self.tag is None:
            raise NotImplementedError("Cannot insert: XML tag not set")
        if isinstance(value, _NewObject):
            raise NotImplementedError("Cannot insert new objects yet")

        self.__create_link(
            elmlist._parent,
            value,
            before=elmlist[index] if index < len(elmlist) else None,
        )

    def delete(
        self,
        elmlist: ElementListCouplingMixin,
        obj: element.ModelObject,
    ) -> None:
        if self.aslist is None:
            raise TypeError("Cannot delete: This is not a list (bug?)")

        parent = elmlist._parent
        for ref in self.__find_refs(parent):
            if self.__follow_ref(parent, ref) == obj._element:
                parent._model._loader.idcache_remove(ref)
                parent._element.remove(ref)
                break
        else:
            raise ValueError("Cannot delete: Target object not in this list")

    @contextlib.contextmanager
    def purge_references(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> cabc.Generator[None, None, None]:
        purge: list[etree._Element] = []
        for i, ref in enumerate(self.__find_refs(obj)):
            if self.__follow_ref(obj, ref) is target._element:
                sys.audit("capellambse.delete", obj, self.__name__, i)
                purge.append(ref)

        yield

        for ref in purge:
            try:
                parent = ref.getparent()
                if parent is None:
                    continue
                parent.remove(ref)
            except Exception:
                LOGGER.exception("Cannot purge dangling ref object %r", ref)


class AttrProxyAccessor(WritableAccessor[T], PhysicalAccessor[T]):
    """Provides access to elements that are linked in an attribute."""

    __slots__ = ("attr",)

    aslist: type[ElementListCouplingMixin] | None
    class_: type[T]

    def __init__(
        self,
        class_: type[T] | None,
        attr: str,
        *,
        aslist: type[element.ElementList] | None = None,
        list_extra_args: cabc.Mapping[str, t.Any] | None = None,
    ):
        """Create an AttrProxyAccessor.

        Parameters
        ----------
        class_
            The proxy class. Currently only used for type hints.
        attr
            The XML attribute to handle.
        aslist
            If None, the attribute contains at most one element
            reference, and either None or the constructed proxy will be
            returned. If not None, must be a subclass of
            :class:`~capellambse.model.common.element.ElementList`. It
            will be used to return a list of all matched objects.
        list_extra_args
            Extra arguments to pass to the
            :class:`~capellambse.model.common.element.ElementList`
            constructor.
        """
        del class_
        super().__init__(
            element.GenericElement,
            aslist=aslist,
            list_extra_args=list_extra_args,
        )
        self.attr = attr

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        elems = obj._model._loader.follow_links(
            obj._element, obj._element.get(self.attr, "")
        )

        rv = self._make_list(obj, elems)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def __set__(
        self,
        obj: element.ModelObject,
        values: T | _NewObject | cabc.Iterable[T | _NewObject],
    ) -> None:
        if getattr(obj, "_constructed", True):
            sys.audit("capellambse.setattr", obj, self.__name__, values)

        if not isinstance(values, cabc.Iterable):
            values = (values,)
        elif self.aslist is None:
            raise TypeError(
                f"{self._qualname} requires a single item, not an iterable"
            )

        if any(isinstance(i, _NewObject) for i in values):
            raise NotImplementedError("Cannot create new objects here")

        assert isinstance(values, cabc.Iterable)
        self.__set_links(obj, values)  # type: ignore[arg-type]

    def __delete__(self, obj: element.ModelObject) -> None:
        if getattr(obj, "_constructed", True):
            sys.audit("capellambse.delete", obj, self.__name__, None)

        del obj._element.attrib[self.attr]

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject | _NewObject,
    ) -> None:
        assert self.aslist is not None
        if isinstance(value, _NewObject):
            raise NotImplementedError("Cannot insert new objects yet")
        if value._model is not elmlist._parent._model:
            raise ValueError("Cannot insert elements from different models")
        objs = [*elmlist[:index], value, *elmlist[index:]]
        self.__set_links(elmlist._parent, objs)

    def delete(
        self, elmlist: ElementListCouplingMixin, obj: element.ModelObject
    ) -> None:
        assert self.aslist is not None
        objs = [i for i in elmlist if i != obj]
        self.__set_links(elmlist._parent, objs)

    def __set_links(
        self, obj: element.ModelObject, values: cabc.Iterable[T]
    ) -> None:
        parts: list[str] = []
        for value in values:
            if value._model is not obj._model:
                raise ValueError(
                    "Cannot insert elements from different models"
                )
            link = obj._model._loader.create_link(obj._element, value._element)
            parts.append(link)
        obj._element.set(self.attr, " ".join(parts))

    @contextlib.contextmanager
    def purge_references(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> cabc.Generator[None, None, None]:
        # pylint: disable=unnecessary-dunder-call
        if self.aslist is not None:
            for i, member in enumerate(self.__get__(obj)):
                if member == target:
                    sys.audit("capellambse.delete", obj, self.__name__, i)
            yield
            try:
                elt = obj._element
                links = obj._model._loader.follow_links(
                    elt, elt.get(self.attr, ""), ignore_broken=True
                )
                remaining_links = [
                    link for link in links if link is not target._element
                ]
                self.__set_links(obj, self._make_list(obj, remaining_links))
            except Exception:
                LOGGER.exception("Cannot write new list of targets")
        else:
            sys.audit("capellambse.delete", obj, self.__name__, None)
            yield
            try:
                del obj._element.attrib[self.attr]
            except KeyError:
                pass
            except Exception:
                LOGGER.exception("Cannot update link target")


class PhysicalLinkEndsAccessor(AttrProxyAccessor[T]):
    def __init__(
        self,
        class_: type[T],
        attr: str,
        *,
        aslist: type[element.ElementList],
    ) -> None:
        super().__init__(
            class_,
            attr,
            aslist=aslist,
            list_extra_args={"fixed_length": 2},
        )
        assert self.aslist is not None

    @contextlib.contextmanager
    def purge_references(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> cabc.Generator[None, None, None]:
        # TODO This should instead delete the link
        raise NotImplementedError("Cannot purge references from PhysicalLink")


class IndexAccessor(Accessor[T]):
    """Access a specific index in an ElementList of a fixed size."""

    __slots__ = ("index", "wrapped")

    def __init__(self, wrapped: str, index: int) -> None:
        super().__init__()
        self.index = index
        self.wrapped = wrapped

    @t.overload
    def __get__(self, obj: None, objtype=None) -> te.Self: ...

    @t.overload
    def __get__(self, obj, objtype=None) -> T | element.ElementList[T]: ...

    def __get__(
        self,
        obj: element.ModelObject | None,
        objtype: type[t.Any] | None = None,
    ) -> te.Self | T | element.ElementList[T]:
        if obj is None:
            return self
        container = getattr(obj, self.wrapped)
        if not isinstance(container, element.ElementList):
            raise RuntimeError(
                f"Cannot get {self._qualname}: {self.wrapped} is not a list"
            )
        if len(container) <= self.index:
            raise RuntimeError(
                f"Broken XML: Expected at least {self.index + 1} elements,"
                f" found {len(container)}"
            )
        return container[self.index]

    def __set__(self, obj: element.ModelObject, value: t.Any) -> None:
        container = getattr(obj, self.wrapped)
        if not isinstance(container, ElementListCouplingMixin):
            raise TypeError(
                f"Cannot set {self._qualname}:"
                f" {self.wrapped} is not a coupled list"
            )
        if len(container) < self.index:
            raise RuntimeError(
                f"Broken XML: Expected at least {self.index+1} elements,"
                f" found {len(container)}"
            )
        container[self.index] = value


class AlternateAccessor(Accessor[T]):
    """Provides access to an "alternate" form of the object."""

    __slots__ = ("class_",)

    def __init__(
        self,
        class_: type[T],
    ):
        super().__init__()
        self.class_ = class_

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self
        rv = self.class_.from_model(obj._model, obj._element)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv


class ParentAccessor(PhysicalAccessor[T]):
    """Accesses the parent XML element."""

    __slots__ = ()

    def __init__(
        self,
        class_: type[T],
    ):
        super().__init__(class_)

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        parent = next(obj._model._loader.iterancestors(obj._element), None)
        if parent is None:
            objrepr = getattr(obj, "_short_repr_", obj.__repr__)()
            raise AttributeError(f"Object {objrepr} is orphaned")
        rv = self.class_.from_model(obj._model, parent)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv


class CustomAccessor(PhysicalAccessor[T]):
    """Customizable alternative to the DirectProxyAccessor.

    .. deprecated:: 0.5.4

       Deprecated due to overcomplexity and (ironically) a lack of
       flexibility.
    """

    __slots__ = (
        "elmfinders",
        "elmmatcher",
        "matchtransform",
    )

    def __init__(
        self,
        class_: type[T],
        *elmfinders: cabc.Callable[[element.GenericElement], cabc.Iterable[T]],
        elmmatcher: cabc.Callable[
            [U, element.GenericElement], bool
        ] = operator.contains,  # type: ignore[assignment]
        matchtransform: cabc.Callable[[T], U] = (
            lambda e: e  # type: ignore[assignment,return-value]
        ),
        aslist: type[element.ElementList] | None = None,
    ) -> None:
        """Create a CustomAccessor.

        Parameters
        ----------
        class_
            The target subclass of ``GenericElement``
        elmfinders
            Functions that are called on the current element. Each
            returns an iterable of possible targets.
        aslist
            If None, only a single element must match, which will be
            returned directly. If not None, must be a subclass of
            :class:`~capellambse.model.common.element.ElementList`,
            which will be used to return a list of all matched objects.
        elmmatcher
            Function that is called with the transformed target element
            and the current element to determine if the untransformed
            target should be accepted.
        matchtransform
            Function that transforms a target so that it can be used by
            the matcher function.
        """
        warnings.warn(
            (
                "CustomAccessor is deprecated,"
                " create a specialized Accessor subclass instead"
            ),
            DeprecationWarning,
            stacklevel=2,
        )

        super().__init__(class_, aslist=aslist)
        self.elmfinders = elmfinders
        self.elmmatcher = elmmatcher
        self.matchtransform = matchtransform

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:
            return self

        elms = itertools.chain.from_iterable(f(obj) for f in self.elmfinders)
        matches = [
            e._element
            for e in elms
            if self.elmmatcher(self.matchtransform(e), obj)
        ]
        return self._make_list(obj, matches)


class AttributeMatcherAccessor(DirectProxyAccessor[T]):
    __slots__ = (
        "_AttributeMatcherAccessor__aslist",
        "attributes",
    )

    def __init__(
        self,
        class_: type[T],
        xtypes: str | type[T] | cabc.Iterable[str | type[T]] | None = None,
        *,
        aslist: type[element.ElementList] | None = None,
        attributes: dict[str, t.Any],
        **kwargs,
    ) -> None:
        super().__init__(
            class_, xtypes, aslist=element.MixedElementList, **kwargs
        )
        self.__aslist = aslist
        self.attributes = attributes

    def __get__(self, obj, objtype=None):
        if obj is None:  # pragma: no cover
            return self

        elements = super().__get__(obj, objtype)
        matches = []
        assert isinstance(elements, cabc.Iterable)
        for elm in elements:
            try:
                if all(
                    getattr(elm, k) == v for k, v in self.attributes.items()
                ):
                    matches.append(elm._element)
            except AttributeError:
                pass

        if self.__aslist is None:
            return no_list(self, obj._model, matches, self.class_)
        return self.__aslist(obj._model, matches, self.class_)


class _Specification(t.MutableMapping[str, str]):
    _aliases = {"LinkedText": "capella:linkedText"}
    _linked_text = frozenset({"capella:linkedText"})
    _model: capellambse.MelodyModel
    _element: etree._Element

    def __init__(
        self, model: capellambse.MelodyModel, elm: etree._Element
    ) -> None:
        self._model = model
        self._element = elm
        self._constructed = True

    def __delitem__(self, k: str) -> None:
        k = self._aliases.get(k, k)
        i, lang_elem = self._index_of(k)
        body_elem = self._body_at(k, i)
        self._element.remove(lang_elem)
        self._element.remove(body_elem)

    def __getitem__(self, k: str) -> str:
        k = self._aliases.get(k, k)
        i, _ = self._index_of(k)
        v = self._body_at(k, i).text or ""
        if k in self._linked_text:
            v = helpers.unescape_linked_text(self._model._loader, v)
        return v

    def __iter__(self) -> cabc.Iterator[str]:
        for i in self._element.iterchildren("languages"):
            yield i.text or ""

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __setitem__(self, k: str, v: str) -> None:
        k = self._aliases.get(k, k)
        if k in self._linked_text:
            v = helpers.escape_linked_text(self._model._loader, v)
        try:
            i, lang = self._index_of(k)
        except KeyError:
            self._element.append(body := self._element.makeelement("bodies"))
            self._element.append(
                lang := self._element.makeelement("languages")
            )
            body.text = v
            lang.text = k
        else:
            body = self._body_at(k, i)
            body.text = v

    def _index_of(self, k: str) -> tuple[int, etree._Element]:
        for i, elm in enumerate(self._element.iterchildren("languages")):
            if elm.text == k:
                return i, elm

        raise KeyError(k)

    def _body_at(self, k: str, i: int) -> etree._Element:
        try:
            return next(
                itertools.islice(
                    self._element.iterchildren("bodies"), i, i + 1
                )
            )
        except StopIteration:
            raise KeyError(k) from None

    def __html__(self) -> str:
        return markupsafe.escape(self.__str__())

    def __str__(self) -> str:  # pragma: no cover
        return next(iter(self.values()))

    def __repr__(self) -> str:  # pragma: no cover
        return f"<{type(self).__name__} at 0x{id(self):016X} {list(self)!r}>"

    @classmethod
    def from_model(cls, _1: capellambse.MelodyModel, _2: t.Any) -> te.Self:
        """Specifications can not be instantiated."""
        raise RuntimeError("Can not create a specification from a model")


class SpecificationAccessor(Accessor[_Specification]):
    """Provides access to linked specifications."""

    __slots__ = ()

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        try:
            spec_elm = next(obj._element.iterchildren("ownedSpecification"))
        except StopIteration:
            raise AttributeError("No specification found") from None

        rv = _Specification(obj._model, spec_elm)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv


class ReferenceSearchingAccessor(PhysicalAccessor[T]):
    """Searches for references to the current element elsewhere."""

    __slots__ = ("attrs", "target_classes")

    attrs: tuple[operator.attrgetter, ...]
    target_classes: tuple[type[element.ModelObject], ...]

    def __init__(
        self,
        class_: type[T] | tuple[type[element.ModelObject], ...],
        *attrs: str,
        aslist: type[element.ElementList] | None = None,
    ) -> None:
        """Create a ReferenceSearchingAccessor.

        Parameters
        ----------
        class_
            The type of class to search for references on.
        attrs
            The attributes of the target classes to search through.
        aslist
            If None, only a single element must match, which will be
            returned directly. If not None, must be a subclass of
            :class:`~capellambse.model.common.element.ElementList`,
            which will be used to return a list of all matched objects.
        """
        if isinstance(class_, tuple):
            super().__init__(
                element.GenericElement,  # type: ignore[arg-type]
                aslist=aslist,
            )
            self.target_classes = class_
        else:
            super().__init__(class_, aslist=aslist)
            self.target_classes = (class_,)
        self.attrs = tuple(operator.attrgetter(i) for i in attrs)

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        matches: list[etree._Element] = []
        for candidate in obj._model.search(*self.target_classes):
            for attr in self.attrs:
                try:
                    value = attr(candidate)
                except AttributeError:
                    continue
                if (
                    isinstance(value, element.ElementList)
                    and obj in value
                    or isinstance(value, element.GenericElement)
                    and obj == value
                ):
                    matches.append(candidate._element)
                    break
        return self._make_list(obj, matches)


class TypecastAccessor(WritableAccessor[T], PhysicalAccessor[T]):
    """Changes the static type of the value of another accessor.

    This is useful for when a class has an attribute that is
    polymorphic, but the accessor should always return a specific
    subclass.

    At runtime, this Accessor mostly behaves like a simple alias
    (without performing any runtime type checks or conversions). When
    creating new objects, it will only allow to create objects of the
    specified type.
    """

    aslist: type[ElementListCouplingMixin] | None
    class_: type[T]

    def __init__(self, cls: type[T], attr: str) -> None:
        super().__init__(cls, (), aslist=element.ElementList)
        self.attr = attr

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> te.Self: ...

    @t.overload
    def __get__(
        self, obj: element.ModelObject, objtype: type[t.Any] | None = None
    ) -> element.ElementList[T]: ...

    def __get__(
        self,
        obj: element.ModelObject | None,
        objtype: type[t.Any] | None = None,
    ) -> te.Self | element.ElementList[T]:
        del objtype
        if obj is None:
            return self

        rv = getattr(obj, self.attr)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def __set__(
        self,
        obj: element.ModelObject,
        value: T | _NewObject | cabc.Iterable[T | _NewObject],
    ) -> None:
        if obj._constructed:
            sys.audit("capellambse.setattr", obj, self.__name__, value)

        if isinstance(value, (list, element.ElementList, tuple)):
            pass
        elif isinstance(value, cabc.Iterable):
            value = list(value)
        else:
            raise TypeError(f"Expected list, got {type(value).__name__}")

        if any(isinstance(i, _NewObject) for i in value):
            raise NotImplementedError("Cannot create new objects here")

        if not all(isinstance(i, self.class_) for i in value):
            raise TypeError(
                f"Expected {self.class_.__name__}, got {type(value).__name__}"
            )
        setattr(obj, self.attr, value)

    def __delete__(self, obj):
        if obj._constructed:
            sys.audit("capellambse.delattr", obj, self.__name__, None)

        delattr(obj, self.attr)

    def create(
        self,
        elmlist: ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> T:
        if type_hints:
            raise TypeError(f"{self._qualname} does not support type hints")
        acc: WritableAccessor = getattr(self.class_, self.attr)
        obj = acc.create(elmlist, build_xtype(self.class_), **kw)
        assert isinstance(obj, self.class_)
        return obj

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject | _NewObject,
    ) -> None:
        if isinstance(value, _NewObject):
            raise NotImplementedError("Cannot insert new objects yet")
        if not isinstance(value, self.class_):
            raise TypeError(
                f"Expected {self.class_.__name__}, got {type(value).__name__}"
            )
        acc: WritableAccessor = getattr(self.class_, self.attr)
        acc.insert(elmlist, index, value)

    def delete(
        self,
        elmlist: ElementListCouplingMixin,
        obj: element.ModelObject,
    ) -> None:
        acc: WritableAccessor = getattr(self.class_, self.attr)
        acc.delete(elmlist, obj)

    @contextlib.contextmanager
    def purge_references(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> cabc.Iterator[None]:
        acc: WritableAccessor = getattr(self.class_, self.attr)
        # XXX: This should really fire a ``capellambse.delete`` audit event,
        #      but the event parameters depend on things only ``acc`` knows.
        with acc.purge_references(obj, target):
            yield


class RoleTagAccessor(WritableAccessor, PhysicalAccessor):
    __slots__ = ("classes", "role_tag")

    aslist: type[ElementListCouplingMixin] | None
    class_: type[element.GenericElement]

    def __init__(
        self,
        role_tag: str,
        classes: (
            type[element.GenericElement]
            | cabc.Iterable[type[element.GenericElement]]
        ) = (),
        *,
        aslist: type[element.ElementList[T]] | None = None,
        list_extra_args: dict[str, t.Any] | None = None,
    ) -> None:
        super().__init__(
            element.GenericElement,
            (),
            aslist=aslist,
            list_extra_args=list_extra_args,
        )
        self.role_tag = role_tag
        if not isinstance(classes, type):
            self.classes = tuple(classes)
        else:
            self.classes = (classes,)

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        elts = list(obj._element.iterchildren(self.role_tag))
        rv = self._make_list(obj, elts)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        if not self.classes:
            return rv
        return rv.filter(lambda i: isinstance(i, self.classes))

    def __set__(
        self,
        obj: element.ModelObject,
        value: (
            str
            | element.GenericElement
            | _NewObject
            | cabc.Iterable[str | element.GenericElement | _NewObject]
        ),
    ) -> None:
        if getattr(obj, "_constructed", True):
            sys.audit("capellambse.setattr", obj, self.__name__, value)

        if self.aslist:
            raise NotImplementedError(
                "Setting lists of model objects is not supported yet"
            )
        else:
            if isinstance(value, cabc.Iterable) and not isinstance(value, str):
                raise TypeError("Cannot set non-list attribute to an iterable")
            if not isinstance(value, _NewObject):
                raise NotImplementedError(
                    "Moving model objects is not supported yet"
                )
            if (elem := self.__get__(obj)) is not None:
                valueclass, _ = self._match_xtype(*value._type_hint)
                elemclass = elem.__class__
                if elemclass is not valueclass:
                    obj._model._loader.idcache_remove(elem._element)
                    obj._element.remove(elem._element)
                else:
                    for k, v in value._kw.items():
                        setattr(elem, k, v)
                    return
            self._create(obj, self.role_tag, *value._type_hint, **value._kw)

    def create(
        self,
        elmlist: ElementListCouplingMixin,
        /,
        *type_hints: str | None,
        **kw: t.Any,
    ) -> element.GenericElement:
        return self._create(elmlist._parent, self.role_tag, *type_hints, **kw)

    def insert(
        self,
        elmlist: ElementListCouplingMixin,
        index: int,
        value: element.ModelObject | _NewObject,
    ) -> None:
        if isinstance(value, _NewObject):
            raise NotImplementedError("Cannot insert new objects yet")
        if value._model is not elmlist._model:
            raise ValueError("Cannot move elements between models")
        try:
            indexof = elmlist._parent._element.index
            if index > 0:
                parent_index = indexof(elmlist._elements[index - 1]) + 1
            elif index < -1:
                parent_index = indexof(elmlist._elements[index + 1]) - 1
            else:
                parent_index = index
        except ValueError:
            parent_index = len(elmlist._parent._element)
        elmlist._parent._element.insert(parent_index, value._element)
        elmlist._model._loader.idcache_index(value._element)

    def delete(
        self,
        elmlist: ElementListCouplingMixin,
        obj: element.ModelObject,
    ) -> None:
        assert obj._model is elmlist._model
        model = obj._model
        all_elements = list(model._loader.iterdescendants_xt(obj._element)) + [
            obj._element
        ]
        with contextlib.ExitStack() as stack:
            for elm in all_elements:
                if elm.get("id") is None:
                    continue

                obj = element.GenericElement.from_model(model, elm)
                for ref, attr, _ in model.find_references(obj):
                    acc = getattr(type(ref), attr)
                    if acc is self or not isinstance(acc, WritableAccessor):
                        continue
                    stack.enter_context(acc.purge_references(ref, obj))

            elm = obj._element
            parent = elm.getparent()
            assert parent is not None
            model._loader.idcache_remove(elm)
            parent.remove(elm)

    @contextlib.contextmanager
    def purge_references(
        self, obj: element.ModelObject, target: element.ModelObject
    ) -> cabc.Iterator[None]:
        yield

    def _match_xtype(
        self,
        type_1: str | None,
        type_2: str | object = _NOT_SPECIFIED,
        /,
    ) -> tuple[type[element.GenericElement], str]:
        if not self.classes:
            return super()._match_xtype(type_1, type_2)

        if type_2 is not _NOT_SPECIFIED:
            raise TypeError(
                f"Only a single type hint is allowed for {self._qualname},"
                f" got ({type_1!r}, {type_2!r})"
            )

        cls = next((i for i in self.classes if i.__name__ == type_1), None)
        if cls is None:
            raise ValueError(f"Invalid class for {self._qualname}: {type_1}")
        return cls, build_xtype(cls)

    def _guess_xtype(self) -> tuple[type[element.GenericElement], str]:
        if len(self.classes) == 1:
            return self.classes[0], build_xtype(self.classes[0])

        if len(self.classes) > 1:
            hint = ", valid values: " + ", ".join(
                i.__name__ for i in self.classes
            )
        else:
            hint = ""
        raise ValueError(f"{self._qualname} requires a type hint{hint}")


def no_list(
    desc: Accessor,
    model: capellambse.MelodyModel,
    elems: cabc.Sequence[etree._Element],
    class_: type[T],
) -> element.ModelObject | None:
    """Return a single element or None instead of a list of elements.

    Parameters
    ----------
    desc
        The descriptor that called this function
    model
        The ``MelodyModel`` instance
    elems
        List of elements that was matched
    class_
        The ``GenericElement`` subclass to instantiate
    """
    if not elems:  # pragma: no cover
        return None
    if len(elems) > 1:  # pragma: no cover
        raise RuntimeError(
            f"Expected 1 object for {desc._qualname}, got {len(elems)}"
        )
    return class_.from_model(model, elems[0])


class ElementListCouplingMixin(element.ElementList[T], t.Generic[T]):
    """Couples an ElementList with an Accessor to enable write support.

    This class is meant to be subclassed further, where the subclass has
    both this class and the originally intended one as base classes (but
    no other ones, i.e. there must be exactly two bases). The Accessor
    then inserts itself as the ``_accessor`` class variable on the new
    subclass. This allows the mixed-in methods to delegate actual model
    modifications to the Accessor.
    """

    _accessor: WritableAccessor[T]

    def __init__(
        self,
        *args: t.Any,
        parent: element.ModelObject,
        fixed_length: int = 0,
        **kw: t.Any,
    ) -> None:
        assert type(self)._accessor

        super().__init__(*args, **kw)
        self._parent = parent
        self.fixed_length = fixed_length

    @t.overload
    def __setitem__(self, index: int, value: T) -> None: ...

    @t.overload
    def __setitem__(self, index: slice, value: cabc.Iterable[T]) -> None: ...

    @t.overload
    def __setitem__(self, index: str, value: t.Any) -> None: ...

    def __setitem__(self, index: int | slice | str, value: t.Any) -> None:
        assert self._parent is not None
        accessor = type(self)._accessor
        assert isinstance(accessor, WritableAccessor)

        if not isinstance(index, (int, slice)):
            super().__setitem__(index, value)
            return

        new_objs = list(self)
        new_objs[index] = value

        if self.fixed_length and len(new_objs) != self.fixed_length:
            raise TypeError(
                f"Cannot set: List must stay at length {self.fixed_length}"
            )

        accessor.__set__(self._parent, new_objs)

    def __delitem__(self, index: int | slice) -> None:
        if self.fixed_length and len(self) <= self.fixed_length:
            raise TypeError("Cannot delete from a fixed-length list")

        assert self._parent is not None
        accessor = type(self)._accessor
        assert isinstance(accessor, WritableAccessor)
        if not isinstance(index, slice):
            index = slice(index, index + 1 or None)
        for obj in self[index]:
            sys.audit(
                "capellambse.delete",
                self._parent,
                accessor.__name__,
                index.start,
            )
            accessor.delete(self, obj)
        super().__delitem__(index)

    def _newlist_type(self) -> type[element.ElementList[T]]:
        assert len(type(self).__bases__) == 2
        assert type(self).__bases__[0] is ElementListCouplingMixin
        return type(self).__bases__[1]

    def create(self, *type_hints: str | None, **kw: t.Any) -> T:
        """Make a new model object (instance of GenericElement).

        Instead of specifying the full ``xsi:type`` including the
        namespace, you can also pass in just the part after the ``:``
        separator. If this is unambiguous, the appropriate
        layer-specific type will be selected automatically.

        This method can be called with or without the ``layertype``
        argument. If a layertype is not given, all layers will be tried
        to find an appropriate ``xsi:type`` handler. Note that setting
        the layertype to ``None`` explicitly is different from not
        specifying it at all; ``None`` tries only the "Transverse
        modelling" type elements.

        Parameters
        ----------
        type_hints
            Hints for finding the correct type of element to create. Can
            either be a full or shortened ``xsi:type`` string, or an
            abbreviation defined by the specific Accessor instance.
        kw
            Initialize the properties of the new object. Depending on
            the object, some attributes may be required.
        """
        if self.fixed_length and len(self) >= self.fixed_length:
            raise TypeError("Cannot create elements in a fixed-length list")

        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, WritableAccessor)
        newobj = acc.create(self, *type_hints, **kw)
        try:
            sys.audit("capellambse.create", self._parent, acc.__name__, newobj)
            acc.insert(self, len(self), newobj)
            super().insert(len(self), newobj)
        except:
            self._parent._element.remove(newobj._element)
            raise
        return newobj

    def create_singleattr(self, arg: t.Any) -> T:
        """Make a new model object (instance of GenericElement).

        This new object has only one interesting attribute.

        See Also
        --------
        :meth:`ElementListCouplingMixin.create` :
            More details on how elements are created.
        :meth:`WritableAccessor.create_singleattr` :
            The method to override in Accessors in order to implement
            this operation.
        """
        if self.fixed_length and len(self) >= self.fixed_length:
            raise TypeError("Cannot create elements in a fixed-length list")

        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, WritableAccessor)
        newobj = acc.create_singleattr(self, arg)
        try:
            sys.audit("capellambse.create", self._parent, acc.__name__, newobj)
            acc.insert(self, len(self), newobj)
            super().insert(len(self), newobj)
        except:
            self._parent._element.remove(newobj._element)
            raise
        return newobj

    def insert(self, index: int, value: T) -> None:
        if self.fixed_length and len(self) >= self.fixed_length:
            raise TypeError("Cannot insert into a fixed-length list")

        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, WritableAccessor)
        if self._parent._constructed:
            sys.audit(
                "capellambse.insert", self._parent, acc.__name__, index, value
            )
        acc.insert(self, index, value)
        super().insert(index, value)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import pytest

import capellambse

# pylint: disable-next=relative-beyond-top-level, useless-suppression
from .conftest import (  # type: ignore[import-untyped]
    INSTALLED_PACKAGE,
    TEST_MODEL,
    TEST_ROOT,
)

TEST_MODEL_AIRD = TEST_ROOT / "5_0" / TEST_MODEL
TEST_MODEL_JSON = INSTALLED_PACKAGE / "known_models" / "test-5.0.json"


def test_enumerate_known_models_contains_known_test_models():
    expected = {"docs.json", "test-5.0.json", "test-5.2.json", "test-lib.json"}

    actual = {i.name for i in capellambse.enumerate_known_models()}

    assert actual >= expected


def test_climodel_is_idempotent(
    session_shared_model: capellambse.MelodyModel,
) -> None:
    paramtype = capellambse.ModelCLI()

    converted = paramtype.convert(session_shared_model, None, None)

    assert converted is session_shared_model


@pytest.mark.parametrize(
    "value",
    [
        pytest.param(TEST_MODEL_JSON.stem, id="str-known"),
        pytest.param(TEST_MODEL_AIRD, id="str-aird"),
        pytest.param(str(TEST_MODEL_AIRD), id="path-aird"),
        pytest.param(TEST_MODEL_JSON, id="str-jsonfile"),
        pytest.param(str(TEST_MODEL_JSON), id="path-jsonfile"),
        pytest.param(TEST_MODEL_JSON.read_text(), id="json"),
    ],
)
def test_climodel_loads_model(value):
    paramtype = capellambse.ModelCLI()

    converted = paramtype.convert(value, None, None)

    assert isinstance(converted, capellambse.MelodyModel)


@pytest.mark.parametrize(
    "value",
    [
        pytest.param(TEST_MODEL_JSON.stem, id="str-known"),
        pytest.param(TEST_MODEL_AIRD, id="str-aird"),
        pytest.param(str(TEST_MODEL_AIRD), id="path-aird"),
        pytest.param(TEST_MODEL_JSON, id="str-jsonfile"),
        pytest.param(str(TEST_MODEL_JSON), id="path-jsonfile"),
        pytest.param(TEST_MODEL_JSON.read_text(), id="json"),
    ],
)
def test_loadcli_loads_model(value):
    model = capellambse.loadcli(value)

    assert isinstance(model, capellambse.MelodyModel)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import itertools
import math
import operator

import pytest

from capellambse import aird, diagram


@pytest.mark.parametrize(
    ["calculate", "vec_1", "vec_2", "expected"],
    [
        # Addition of two vectors
        (
            operator.add,
            diagram.Vector2D(2, 5),
            diagram.Vector2D(8, 1),
            diagram.Vector2D(10, 6),
        ),
        (
            operator.add,
            diagram.Vector2D(2, 5),
            (8, 1),
            diagram.Vector2D(10, 6),
        ),
        (
            operator.add,
            (2, 5),
            diagram.Vector2D(8, 1),
            diagram.Vector2D(10, 6),
        ),
        # Subtraction of two vectors
        (
            operator.sub,
            diagram.Vector2D(8, 5),
            diagram.Vector2D(2, 1),
            diagram.Vector2D(6, 4),
        ),
        (operator.sub, diagram.Vector2D(8, 5), (2, 1), diagram.Vector2D(6, 4)),
        (operator.sub, (8, 5), diagram.Vector2D(2, 1), diagram.Vector2D(6, 4)),
        # Multiplication with scalar
        (operator.mul, diagram.Vector2D(3, 4), 2, diagram.Vector2D(6, 8)),
        (operator.mul, 2, diagram.Vector2D(3, 4), diagram.Vector2D(6, 8)),
        # Multiplication with vector
        (operator.mul, diagram.Vector2D(3, 4), diagram.Vector2D(5, 2), 23),
        (operator.mul, diagram.Vector2D(3, 4), (5, 2), 23),
        (operator.mul, (3, 4), diagram.Vector2D(5, 2), 23),
        # Matrix multiplication
        (
            operator.matmul,
            diagram.Vector2D(3, 4),
            diagram.Vector2D(5, 2),
            diagram.Vector2D(15, 8),
        ),
        (
            operator.matmul,
            diagram.Vector2D(3, 4),
            (5, 2),
            diagram.Vector2D(15, 8),
        ),
        (
            operator.matmul,
            (3, 4),
            diagram.Vector2D(5, 2),
            diagram.Vector2D(15, 8),
        ),
        # True division
        (
            operator.truediv,
            diagram.Vector2D(8, 6),
            2,
            diagram.Vector2D(4.0, 3.0),
        ),
        (
            operator.truediv,
            8,
            diagram.Vector2D(4, 2),
            diagram.Vector2D(2.0, 4.0),
        ),
        # Floor division
        (operator.floordiv, diagram.Vector2D(8, 6), 2, diagram.Vector2D(4, 3)),
        (operator.floordiv, 8, diagram.Vector2D(4, 2), diagram.Vector2D(2, 4)),
    ],
)
def test_vector_math(calculate, vec_1, vec_2, expected):
    actual = calculate(vec_1, vec_2)
    assert isinstance(actual, type(expected))
    if isinstance(expected, diagram.Vector2D):
        assert isinstance(actual[0], type(expected[0]))
        assert isinstance(actual[1], type(expected[1]))
    assert actual == expected


@pytest.mark.parametrize(
    ["vector", "expected"],
    [
        (diagram.Vector2D(-2, 1), (2, 1)),
        (diagram.Vector2D(3, -5), (3, 5)),
        (diagram.Vector2D(2, 2), (2, 2)),
        (diagram.Vector2D(-7, -4), (7, 4)),
    ],
)
def test_abs_makes_negative_components_positive(vector, expected):
    actual = abs(vector)
    assert isinstance(actual, diagram.Vector2D)
    assert actual == expected


@pytest.mark.parametrize(
    ["vector", "expected"],
    [
        (diagram.Vector2D(-2, 1), 5),
        (diagram.Vector2D(3, -5), 34),
        (diagram.Vector2D(2, 2), 8),
        (diagram.Vector2D(-7, -4), 65),
    ],
)
def test_squared_length(vector, expected):
    actual = vector.sqlength
    assert math.isclose(actual, expected)


@pytest.mark.parametrize(
    ["vector", "expected"],
    [
        (diagram.Vector2D(-2, 1), 2.23606797749979),
        (diagram.Vector2D(3, -5), 5.830951894845301),
        (diagram.Vector2D(2, 2), 2.8284271247461903),
        (diagram.Vector2D(-7, -4), 8.06225774829855),
    ],
)
def test_length(vector, expected):
    actual = vector.length
    assert math.isclose(actual, expected)


@pytest.mark.parametrize(
    ["vector", "expected"],
    [
        (diagram.Vector2D(-2, 1), (-0.8944271909999159, 0.4472135954999579)),
        (diagram.Vector2D(3, -5), (0.5144957554275265, -0.8574929257125441)),
        (diagram.Vector2D(2, 2), (0.7071067811865475, 0.7071067811865475)),
        (
            diagram.Vector2D(-7, -4),
            (-0.8682431421244593, -0.49613893835683387),
        ),
    ],
)
def test_normalized(vector, expected):
    actual = vector.normalized
    assert math.isclose(actual[0], expected[0])
    assert math.isclose(actual[1], expected[1])


@pytest.mark.parametrize(
    ["vector", "expected"],
    [
        # Exactly on the box - noop
        (diagram.Vector2D(1, 1), diagram.Vector2D(1, 1)),
        (diagram.Vector2D(1, 3), diagram.Vector2D(1, 3)),
        (diagram.Vector2D(5, 1), diagram.Vector2D(5, 1)),
        (diagram.Vector2D(5, 3), diagram.Vector2D(5, 3)),
        (diagram.Vector2D(1, 2), diagram.Vector2D(1, 2)),
        (diagram.Vector2D(5, 2), diagram.Vector2D(5, 2)),
        (diagram.Vector2D(3, 1), diagram.Vector2D(3, 1)),
        (diagram.Vector2D(3, 3), diagram.Vector2D(3, 3)),
        # Inside the box
        (diagram.Vector2D(3, 1.5), diagram.Vector2D(3, 1)),
        (diagram.Vector2D(1.5, 2), diagram.Vector2D(1, 2)),
        (diagram.Vector2D(3, 2.5), diagram.Vector2D(3, 3)),
        (diagram.Vector2D(4.5, 2), diagram.Vector2D(5, 2)),
        # Outside left
        (diagram.Vector2D(0, 2), diagram.Vector2D(1, 2)),
        # Outside right
        (diagram.Vector2D(6, 2), diagram.Vector2D(5, 2)),
        # Outside top
        (diagram.Vector2D(3, 0), diagram.Vector2D(3, 1)),
        # Outside bottom
        (diagram.Vector2D(3, 4), diagram.Vector2D(3, 3)),
        # Outside top-left
        (diagram.Vector2D(0, 0), diagram.Vector2D(1, 1)),
        (diagram.Vector2D(-1, 0), diagram.Vector2D(1, 1)),
        (diagram.Vector2D(0, -1), diagram.Vector2D(1, 1)),
        # Outside top-right
        (diagram.Vector2D(6, 0), diagram.Vector2D(5, 1)),
        (diagram.Vector2D(7, 0), diagram.Vector2D(5, 1)),
        (diagram.Vector2D(6, -1), diagram.Vector2D(5, 1)),
        # Outside bottom-left
        (diagram.Vector2D(0, 4), diagram.Vector2D(1, 3)),
        (diagram.Vector2D(-1, 4), diagram.Vector2D(1, 3)),
        (diagram.Vector2D(0, 5), diagram.Vector2D(1, 3)),
        # Outside bottom-right
        (diagram.Vector2D(6, 4), diagram.Vector2D(5, 3)),
        (diagram.Vector2D(7, 4), diagram.Vector2D(5, 3)),
        (diagram.Vector2D(6, 5), diagram.Vector2D(5, 3)),
    ],
)
def test_boxsnap(vector: diagram.Vector2D, expected: diagram.Vector2D):
    topleft = diagram.Vector2D(1, 1)
    botright = diagram.Vector2D(5, 3)

    actual = vector.boxsnap(topleft, botright)

    assert actual == expected


@pytest.mark.parametrize(
    ["vector1", "vector2", "expected"],
    [
        pytest.param(
            (diagram.Vector2D(0, 0), diagram.Vector2D(1, 1)),
            (diagram.Vector2D(1, 0), diagram.Vector2D(0, 1)),
            diagram.Vector2D(0.5, 0.5),
            id="Simple",
        ),
        pytest.param(
            (diagram.Vector2D(1, 0), diagram.Vector2D(2, 0)),
            (diagram.Vector2D(0, 6), diagram.Vector2D(0, 3)),
            diagram.Vector2D(0, 0),
            id="Intersection not between given points",
        ),
    ],
)
def test_line_intersect(
    vector1: tuple[diagram.Vector2D, diagram.Vector2D],
    vector2: tuple[diagram.Vector2D, diagram.Vector2D],
    expected: diagram.Vector2D,
):
    point = diagram.line_intersect(vector1, vector2)

    assert point == expected


def test_line_intersect_raises_ValueError_when_parallel_lines_given():
    with pytest.raises(ValueError):
        diagram.line_intersect(
            (diagram.Vector2D(0, 0), diagram.Vector2D(0, 1)),
            (diagram.Vector2D(1, 0), diagram.Vector2D(1, 1)),
        )


class TestVectorSnapping:
    PORTBOX = diagram.Box((1, 1), aird._common.PORT_SIZE, port=True)
    PORT_WEST = diagram.Vector2D(1, 6)
    PORT_NORTH = diagram.Vector2D(6, 1)
    PORT_EAST = diagram.Vector2D(11, 6)
    PORT_SOUTH = diagram.Vector2D(6, 11)

    @pytest.mark.parametrize(
        ["points", "expected"],
        [
            pytest.param(
                [diagram.Vector2D(0, -1), diagram.Vector2D(0, 1)],
                PORT_WEST,
                id="West - Outside - Parallel(above)",
            ),
            pytest.param(
                [diagram.Vector2D(-1, 1), diagram.Vector2D(0, 1)],
                PORT_WEST,
                id="West - Outside - Perpendicular",
            ),
            pytest.param(
                [diagram.Vector2D(0, 2), diagram.Vector2D(0, 1)],
                PORT_WEST,
                id="West - Outside - Parallel(lower)",
            ),
            pytest.param(
                [diagram.Vector2D(0, 5), diagram.Vector2D(0, 6)],
                PORT_WEST,
                id="West - Outside - Parallel(above) - Perfect align",
            ),
            pytest.param(
                [diagram.Vector2D(-1, 6), diagram.Vector2D(0, 6)],
                PORT_WEST,
                id="West - Outside - Perpendicular - Perfect align",
            ),
            pytest.param(
                [diagram.Vector2D(0, 7), diagram.Vector2D(0, 6)],
                PORT_WEST,
                id="West - Outside - Parallel(lower) - Perfect align",
            ),
            pytest.param(
                [diagram.Vector2D(0, 10), diagram.Vector2D(0, 11)],
                PORT_WEST,
                id="West - Outside - Parallel(above)",
            ),
            pytest.param(
                [diagram.Vector2D(-1, 11), diagram.Vector2D(0, 11)],
                PORT_WEST,
                id="West - Outside - Perpendicular",
            ),
            pytest.param(
                [diagram.Vector2D(0, 12), diagram.Vector2D(0, 11)],
                PORT_WEST,
                id="West - Outside - Parallel(lower)",
            ),
        ],
    )
    def test_snap_manhattan_to_middle_of_port_side(
        self, points: list[diagram.Vector2D], expected: diagram.Vector2D
    ):
        aird._edge_factories.snaptarget(
            points, -1, -2, self.PORTBOX, False, "manhattan"
        )

        points_ = itertools.tee(points, 2)
        next(points_[1], None)
        for p, q in zip(*points_):
            direction = p - q
            assert direction and (direction.x == 0 or direction.y == 0)
        assert points[-1] == expected


@pytest.mark.parametrize(
    ["point", "source", "expected"],
    [
        # horizontal
        ((20, -5), (19, -5), (5, 0)),
        ((-5, 5), (-6, 5), (0, 5)),
        ((-5, 7), (-6, 7), (0, 5)),
        ((20, 15), (19, 15), (5, 10)),
        ((20, 5), (19, 5), (0, 5)),
        ((20, 2), (19, 2), (0, 5)),
        # vertical
        ((-5, 20), (-5, 19), (0, 5)),
        ((5, -5), (5, -6), (5, 0)),
        ((7, -5), (7, -6), (5, 0)),
        ((15, 20), (15, 19), (10, 5)),
        ((5, 20), (5, 19), (5, 0)),
        ((2, 20), (2, 19), (5, 0)),
    ],
)
def test_box_snapping_manhattan(
    point: diagram.Vec2ish, source: diagram.Vec2ish, expected: diagram.Vec2ish
) -> None:
    box = diagram.Box((0, 0), (10, 10), port=True)
    style = diagram.RoutingStyle.MANHATTAN

    new_point = box.vector_snap(point, source=source, style=style)

    assert new_point == expected


@pytest.mark.parametrize(
    ["point", "source", "expected"],
    [
        # horizontal
        ((2, 0), (-4, 0), (2, 0)),
        ((2, 10), (-4, 10), (2, 10)),
        ((2, 0), (22, 0), (2, 0)),
        ((2, 10), (22, 10), (2, 10)),
        # vertical
        ((4, 20), (4, 40), (4, 10)),
        ((4, -10), (4, -40), (4, 0)),
    ],
)
def test_box_snapping_tree(
    point: diagram.Vec2ish, source: diagram.Vec2ish, expected: diagram.Vec2ish
) -> None:
    box = diagram.Box((0, 0), (10, 10), port=False)
    style = diagram.RoutingStyle.TREE

    new_point = box.vector_snap(point, source=source, style=style)

    assert new_point == expected


@pytest.mark.parametrize(
    ["point", "source", "expected"],
    [
        # horizontal
        ((2, 0), (-4, 0), (5, 0)),
        ((2, 10), (-4, 10), (5, 10)),
        ((2, 0), (22, 0), (5, 0)),
        ((2, 10), (22, 10), (5, 10)),
        # vertical
        ((4, 20), (4, 40), (5, 10)),
        ((4, -10), (4, -40), (5, 0)),
    ],
)
def test_box_port_snapping_tree(
    point: diagram.Vec2ish, source: diagram.Vec2ish, expected: diagram.Vec2ish
) -> None:
    box = diagram.Box((0, 0), (10, 10), port=True)
    style = diagram.RoutingStyle.TREE

    new_point = box.vector_snap(point, source=source, style=style)

    assert new_point == expected



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Enumeration types used by the MelodyModel."""
import enum as _enum
import typing as t


class _StringyEnumMixin:
    """Mixin for enums that makes members compare equal to their key name.

    :meta public:
    """

    name: t.Any
    _name_: t.Any

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self is other
        if isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __str__(self) -> str:
        return str(self.name)

    def __hash__(self):
        return hash(self._name_)


class DiagramType(_StringyEnumMixin, _enum.Enum):
    """The types of diagrams that Capella knows about.

    Extracted from::

        $CAPELLA/eclipse/configuration/org.eclipse.osgi/635/0/.cp/description

    with::

        grep '<ownedRepresentations' *(.) \
        | grep --color=always -P '(?<=name=").*?(?=")'
    """

    UNKNOWN = "(Unknown Diagram Type)"
    # Common
    MSM = "Mode State Machine"
    # Capella Architecture?
    CDI = "Contextual Component Detailed Interfaces"
    CEI = "Contextual Component External Interfaces"
    CII = "Contextual Component Internal Interfaces"
    IDB = "Interfaces Diagram Blank"
    # Requirements?
    CRI = "Contextual Capability Realization Involvement"
    CRB = "Capability Realization Blank"
    PD = "Package Dependencies"
    ID = "Interface Delegations"
    CDB = "Class Diagram Blank"
    IS = "Component Interfaces Scenario"
    ES = "Component Exchanges Scenario"
    FS = "Functional Scenario"
    SFCD = LFCD = PFCD = "Functional Chain Description"
    # State And Mode - Matrix?
    # Contextual State And Mode - Matrix?
    # Modes and States Reference Matrix?
    # Operational Analysis
    # Operational Activities - Requirements?
    OEBD = "Operational Entity Breakdown"
    OAIB = "Operational Activity Interaction Blank"
    OAB = "Operational Entity Blank"
    OABD = "Operational Activity Breakdown"
    ORB = "Operational Role Blank"
    OES = "Operational Interaction Scenario"
    OAS = "Activity Interaction Scenario"
    OPD = "Operational Process Description"
    OCB = "Operational Capabilities Blank"
    # Requirements - Operational Activities?
    COC = "Contextual Operational Capability"
    # System Analysis
    CM = "Contextual Mission"
    MB = "Missions Blank"
    CC = "Contextual Capability"
    MCB = "Missions Capabilities Blank"
    # System Functions - Requirements?
    # System Functions - Operational Activities?
    SFBD = "System Function Breakdown"
    SDFB = "System Data Flow Blank"
    SAB = "System Architecture Blank"
    CSA = "Contextual System Actors"
    # System Actor - Operational Actor?
    # Interfaces - Capabilities?
    # Interfaces - Scenarios?
    # Interfaces - Capabilities and Scenarios?
    # System/Actors - System Functions?
    # Requirements - System Functions?
    # Logical Architecture
    # Logical Functions - Requirements?
    # Logical Components - Requirements?
    # Logical Functions - System Functions?
    # Logical Components - Logical Functions?
    # Logical Architecture Requirement Refinements?
    # Logical Interface - Context Interface?
    # Logical Actor - Context Actor?
    LCBD = "Logical Component Breakdown"
    LFBD = "Logical Function Breakdown"
    LDFB = "Logical Data Flow Blank"
    LAB = "Logical Architecture Blank"
    CRR = "Capability Realization Refinement"
    # Requirements - Logical Functions?
    # Physical Architecture
    # Physical Functions - Requirements?
    # Physical Components - Requirements?
    # Physical Functions - Logical Functions?
    # Physical Components - Logical Components?
    # Physical Components - Physical Functions?
    # Physical Interface - Logical Interface?
    PFBD = "Physical Function Breakdown"
    PDFB = "Physical Data Flow Blank"
    PCBD = "Physical Component Breakdown"
    PAB = "Physical Architecture Blank"
    # Physical Actor - Logical Actor?
    # Requirements - Physical Functions?
    PPD = "Physical Path Description"
    # EPBS
    # Configuration Items - Requirements?
    # Configuration Items - Physical Artifacts?
    # EPBS Requirement Refinements?
    EAB = "EPBS Architecture Blank"
    CIBD = "Configuration Items Breakdown"


@_enum.unique
class AccessPolicy(_StringyEnumMixin, _enum.Enum):
    READ_ONLY = "readOnly"
    READ_AND_WRITE = "readAndWrite"


@_enum.unique
class AggregationKind(_StringyEnumMixin, _enum.Enum):
    """Defines the specific kind of a relationship, as per UML definitions."""

    UNSET = "UNSET"
    """Used when value is not defined by the user."""
    ASSOCIATION = "ASSOCIATION"
    """A semantic relationship between typed instances.

    It has at least two ends represented by properties, each of which is
    connected to the type of the end. More than one end of the
    association may have the same type.

    Indicates that the property has no aggregation.
    """
    AGGREGATION = "AGGREGATION"
    """A semantic relationship between a part and a whole.

    The part has a lifecycle of its own, and is potentially shared among
    several aggregators.
    """
    COMPOSITION = "COMPOSITION"
    """A semantic relationship between whole and its parts.

    The parts lifecycles are tied to that of the whole, and they are not
    shared with any other aggregator.
    """


@_enum.unique
class BinaryOperator(_StringyEnumMixin, _enum.Enum):
    """Specifies the kind of this binary operator."""

    UNSET = "UNSET"
    """The binary operator is not initialized."""
    ADD = "ADD"
    """The binary operator refers to an addition."""
    MUL = "MUL"
    """The binary operator refers to a multiplication."""
    SUB = "SUB"
    """The binary operator refers to a substraction."""
    DIV = "DIV"
    """The binary operator refers to a division."""
    POW = "POW"
    """The binary operator refers to a power operation."""
    MIN = "MIN"
    """The binary operator refers to a min operation."""
    MAX = "MAX"
    """The binary operator refers to a max operation."""
    EQU = "EQU"
    """The binary operator refers to an equal operation."""
    IOR = "IOR"
    """The binary operator refers to a logical inclusive OR operation."""
    XOR = "XOR"
    """The binary operator refers to a logical exclusive OR operation."""
    AND = "AND"
    """The binary operator refers to a logical AND operation."""


@_enum.unique
class CatalogElementKind(_StringyEnumMixin, _enum.Enum):
    REC = "REC"
    RPL = "RPL"
    REC_RPL = "REC_RPL"
    GROUPING = "GROUPING"


@_enum.unique
class ChangeEventKind(_StringyEnumMixin, _enum.Enum):
    WHEN = "WHEN"


@_enum.unique
class CollectionKind(_StringyEnumMixin, _enum.Enum):
    """Defines the specific kind of a Collection structure."""

    ARRAY = "ARRAY"
    """The collection is to be considered an array of elements."""
    SEQUENCE = "SEQUENCE"
    """The collection is to be considered as a sequence (list) of elements."""


@_enum.unique
class CommunicationLinkKind(_StringyEnumMixin, _enum.Enum):
    """Enumeration listing the various possibilities of communication links."""

    UNSET = "UNSET"
    """The CommunicationLink protocol is not yet set."""
    PRODUCE = "PRODUCE"
    """The CommunicationLink describes a production of ExchangeItem."""
    CONSUME = "CONSUME"
    """The CommunicationLink describes a comsumption of ExchangeItem."""
    SEND = "SEND"
    """The CommunicationLink describes a sending of ExchangeItem."""
    RECEIVE = "RECEIVE"
    """The CommunicationLink describes a reception of ExchangeItem."""
    CALL = "CALL"
    """The CommunicationLink describes a call of ExchangeItem."""
    EXECUTE = "EXECUTE"
    """The CommunicationLink describes an execution of ExchangeItem."""
    WRITE = "WRITE"
    """The CommunicationLink describes a writing of ExchangeItem."""
    ACCESS = "ACCESS"
    """The CommunicationLink describes an access to the ExchangeItem."""
    ACQUIRE = "ACQUIRE"
    """The CommunicationLink describes an acquisition of ExchangeItem."""
    TRANSMIT = "TRANSMIT"
    """The CommunicationLink describes a transmission of ExchangeItem."""


@_enum.unique
class CommunicationLinkProtocol(_StringyEnumMixin, _enum.Enum):
    """The various possibilities for the protocol of the communication link."""

    UNSET = "UNSET"
    """The CommunicationLink protocol is not yet set."""
    UNICAST = "UNICAST"
    """Describes sending an ExchangeItem using the unicast protocol."""
    MULTICAST = "MULTICAST"
    """Describes sending an ExchangeItem using the multicast protocol."""
    BROADCAST = "BROADCAST"
    """Describes sending an ExchangeItem using the broadcast protocol."""
    SYNCHRONOUS = "SYNCHRONOUS"
    """Describes a call of the ExchangeItem using the synchronous protocol."""
    ASYNCHRONOUS = "ASYNCHRONOUS"
    """Describes a call of the ExchangeItem using the asynchronous protocol."""
    READ = "READ"
    """Describes access to the ExchangeItem by reading it."""
    ACCEPT = "ACCEPT"
    """Describes access to the ExchangeItem by accepting it."""


@_enum.unique
class ComponentExchangeKind(_StringyEnumMixin, _enum.Enum):
    """The kind of a ComponentExchange."""

    UNSET = "UNSET"
    """Communication kind is not set."""
    DELEGATION = "DELEGATION"
    """Indicates that the connector is a delegation connector."""
    ASSEMBLY = "ASSEMBLY"
    """Indicates that the connector is an assembly connector."""
    FLOW = "FLOW"
    """Describes a flow communication."""


@_enum.unique
class ComponentPortKind(_StringyEnumMixin, _enum.Enum):
    STANDARD = "STANDARD"
    """Describes a standard port.

    A port is an interaction point between a Block or sub-Block and its
    environment that supports Exchanges with other ports.
    """
    FLOW = "FLOW"
    """Describes a flow port.

    A flow port is an interaction point through which input and/or
    output of items such as data, material, or energy may flow.
    """


@_enum.unique
class ConfigurationItemKind(_StringyEnumMixin, _enum.Enum):
    UNSET = "Unset"
    COTSCI = "COTSCI"
    """Commercial Off The Shelves Configuration Item."""
    CSCI = "CSCI"
    """Computer Software Configuration Item."""
    HWCI = "HWCI"
    """Hardware Configuration Item."""
    INTERFACE_CI = "InterfaceCI"
    """Interface Configuration Item."""
    NDICI = "NDICI"
    """Non Developmental Configuration Item."""
    PRIME_ITEM_CI = "PrimeItemCI"
    """Prime Item Configuration Item."""
    SYSTEM_CI = "SystemCI"
    """System Configuration Item."""


@_enum.unique
class ControlNodeKind(_StringyEnumMixin, _enum.Enum):
    OR = "OR"
    AND = "AND"
    ITERATE = "ITERATE"


@_enum.unique
class ElementKind(_StringyEnumMixin, _enum.Enum):
    """The visibility options for features of a class."""

    TYPE = "TYPE"
    """The ExchangeItemElement is a type for its ExchangeItem."""
    MEMBER = "MEMBER"
    """The ExchangeItemElement is a member for its ExchangeItem."""


@_enum.unique
class ExchangeMechanism(_StringyEnumMixin, _enum.Enum):
    """Enumeration of the different exchange mechanisms."""

    UNSET = "UNSET"
    """The exchange mechanism is not defined."""
    FLOW = "FLOW"
    """Continuous supply of data."""
    OPERATION = "OPERATION"
    """Sporadic supply of data with returned data."""
    EVENT = "EVENT"
    """Asynchronous information that is taken into account rapidly."""
    SHARED_DATA = "SHARED_DATA"


@_enum.unique
class FunctionalChainKind(_StringyEnumMixin, _enum.Enum):
    """The kind of a Functional Chain."""

    SIMPLE = "SIMPLE"
    COMPOSITE = "COMPOSITE"
    FRAGMENT = "FRAGMENT"


@_enum.unique
class FunctionKind(_StringyEnumMixin, _enum.Enum):
    """The kind of a Function."""

    FUNCTION = "FUNCTION"
    DUPLICATE = "DUPLICATE"
    GATHER = "GATHER"
    SELECT = "SELECT"
    SPLIT = "SPLIT"
    ROUTE = "ROUTE"


@_enum.unique
class InteractionOperatorKind(_StringyEnumMixin, _enum.Enum):
    UNSET = "UNSET"
    ALT = "ALT"
    OPT = "OPT"
    PAR = "PAR"
    LOOP = "LOOP"
    CRITICAL = "CRITICAL"
    NEG = "NEG"
    ASSERT = "ASSERT"
    STRICT = "STRICT"
    SEQ = "SEQ"
    IGNORE = "IGNORE"
    CONSIDER = "CONSIDER"


@_enum.unique
class MessageKind(_StringyEnumMixin, _enum.Enum):
    """Identifies the type of message.

    This concept is similar to UML ``MessageSort``.
    """

    UNSET = "UNSET"
    """The message kind is not specified."""
    ASYNCHRONOUS_CALL = "ASYNCHRONOUS_CALL"
    """The message was generated by an asynchronous call to an operation.

    Equivalent to UML ``MessageSort::asynchCall``.
    """
    SYNCHRONOUS_CALL = "SYNCHRONOUS_CALL"
    """The message was generated by a synchronous call to an operation.

    Equivalent to UML ``MessageSort::synchCall``.
    """
    REPLY = "REPLY"
    """The message is a reply message to an operation call.

    Equivalent to UML ``MessageSort::reply``.
    """
    DELETE = "DELETE"
    """The message designates the termination of another lifeline.

    Equivalent to UML ``MessageSort::deleteMessage``.
    """
    CREATE = "CREATE"
    """The message designates the creation of an instance role."""
    TIMER = "TIMER"


@_enum.unique
class NumericTypeKind(_StringyEnumMixin, _enum.Enum):
    """The kind of this numeric data type."""

    INTEGER = "INTEGER"
    FLOAT = "FLOAT"


@_enum.unique
class ObjectNodeKind(_StringyEnumMixin, _enum.Enum):
    """The behavior type of the object node with respect to incoming data."""

    UNSPECIFIED = "Unspecified"
    """Used when incoming object node management policy is not specified."""
    NO_BUFFER = "NoBuffer"
    """Discard incoming tokens if they are refused.

    When the "nobuffer" stereotype is applied to object nodes, tokens
    arriving at the node are discarded if they are refused by outgoing
    edges, or refused by actions for object nodes that are input pins.
    """
    OVERWRITE = "Overwrite"
    """Incoming tokens may overwrite existing ones.

    When the "overwrite" stereotype is applied to object nodes, a token
    arriving at a full object node replaces the ones already there. A
    full object node has as many tokens as allowed by its upper bound.
    """


@_enum.unique
class ObjectNodeOrderingKind(_StringyEnumMixin, _enum.Enum):
    """Indicates queuing order within a node."""

    FIFO = "FIFO"
    """First In First Out ordering."""
    LIFO = "LIFO"
    """Last In First Out ordering."""
    ORDERED = "ordered"
    """Indicates that object node tokens are ordered."""
    UNORDERED = "unordered"
    """Indicates that object node tokens are unordered."""


@_enum.unique
class OrientationPortKind(_StringyEnumMixin, _enum.Enum):
    """Direction of component ports."""

    UNSET = "UNSET"
    """The port orientation is undefined."""
    IN = "IN"
    """The port represents an input of the component it is used in."""
    OUT = "OUT"
    """The port represents an output of the component it is used in."""
    INOUT = "INOUT"
    """The port represents both an input and on output of the component."""


@_enum.unique
class ParameterDirection(_StringyEnumMixin, _enum.Enum):
    """The direction in which data is passed along through a parameter."""

    IN = "IN"
    """The parameter represents an input of the operation it is used in."""
    OUT = "OUT"
    """The parameter represents an output of the operation it is used in."""
    INOUT = "INOUT"
    """The parameter represents both an input and output of the operation."""
    RETURN = "RETURN"
    """The parameter represents the return value of the operation."""
    EXCEPTION = "EXCEPTION"
    """The parameter is like an exception."""
    UNSET = "UNSET"
    """The CommunicationLink protocol is not yet set."""


@_enum.unique
class ParameterEffectKind(_StringyEnumMixin, _enum.Enum):
    """A behavior's effect on values passed in or out of its parameters."""

    CREATE = "create"
    """The parameter value is being created upon behavior execution."""
    READ = "read"
    """The parameter value is only being read upon behavior execution."""
    UPDATE = "update"
    """The parameter value is being updated upon behavior execution."""
    DELETE = "delete"
    """The parameter value is being deleted upon behavior execution."""


@_enum.unique
class PassingMode(_StringyEnumMixin, _enum.Enum):
    """The data passing mechanism for parameters of an operation."""

    UNSET = "UNSET"
    """The data passing mechanism is not precised."""
    BY_REF = "BY_REF"
    """The data is being passed by reference to the operation."""
    BY_VALUE = "BY_VALUE"
    """The data is being passed by value to the operation."""


@_enum.unique
class PhysicalComponentKind(_StringyEnumMixin, _enum.Enum):
    """Categories of physical components.

    Allows to categorize a physical component, with respect to real life
    physical entities.
    """

    UNSET = "UNSET"
    """The physical component kind is not specified."""
    HARDWARE = "HARDWARE"
    """The physical component is a hardware resource."""
    HARDWARE_COMPUTER = "HARDWARE_COMPUTER"
    """The physical component is a computing resource."""
    SOFTWARE = "SOFTWARE"
    """The physical component is a software entity."""
    SOFTWARE_DEPLOYMENT_UNIT = "SOFTWARE_DEPLOYMENT_UNIT"
    """The physical component is a software deployment unit."""
    SOFTWARE_EXECUTION_UNIT = "SOFTWARE_EXECUTION_UNIT"
    """The physical component is a software execution unit."""
    SOFTWARE_APPLICATION = "SOFTWARE_APPLICATION"
    """The physical component is a software application."""
    FIRMWARE = "FIRMWARE"
    """The physical component is a firmware part."""
    PERSON = "PERSON"
    """The physical component is a person."""
    FACILITIES = "FACILITIES"
    """The physical component refers to Facilities."""
    DATA = "DATA"
    """The physical component represents a set of data."""
    MATERIALS = "MATERIALS"
    """The physical component represents a bunch of materials."""
    SERVICES = "SERVICES"
    """The physical component represents a set of services."""
    PROCESSES = "PROCESSES"
    """The physical component represents a set of processes."""


@_enum.unique
class PhysicalComponentNature(_StringyEnumMixin, _enum.Enum):
    """The nature of a physical component."""

    UNSET = "UNSET"
    """The physical component nature is not specified."""
    BEHAVIOR = "BEHAVIOR"
    """The physical component nature is behavioral.

    This typically means a piece of software.
    """
    NODE = "NODE"
    """The physical component is a host for behavioral components.

    This typically means a computing resource.
    """


@_enum.unique
class RateKind(_StringyEnumMixin, _enum.Enum):
    """The possible caracterizations for the rate of a streaming parameter."""

    UNSPECIFIED = "Unspecified"
    """The rate kind is not specified."""
    CONTINUOUS = "Continuous"
    """The rate characterizes a continuous flow."""
    DISCRETE = "Discrete"
    """The rate characterizes a discrete flow."""


@_enum.unique
class ScenarioKind(_StringyEnumMixin, _enum.Enum):
    UNSET = "UNSET"
    INTERFACE = "INTERFACE"
    DATA_FLOW = "DATA_FLOW"
    INTERACTION = "INTERACTION"
    FUNCTIONAL = "FUNCTIONAL"


@_enum.unique
class SynchronismKind(_StringyEnumMixin, _enum.Enum):
    """The synchronicity of an operation invocation."""

    UNSET = "UNSET"
    SYNCHRONOUS = "SYNCHRONOUS"
    ASYNCHRONOUS = "ASYNCHRONOUS"


@_enum.unique
class TimeEventKind(_StringyEnumMixin, _enum.Enum):
    AT = "AT"
    """Trigger at a specific time.

    An absolute time trigger is specified with the keyword 'at' followed
    by an expression that evaluates to a time value, such as 'Jan. 1,
    2000, Noon'.
    """
    AFTER = "AFTER"
    """Trigger after a relative time duration has passed.

    A relative time trigger is specified with the keyword 'after'
    followed by an expression that evaluates to a time value, such as
    'after (5 seconds)'.
    """


@_enum.unique
class TransitionKind(_StringyEnumMixin, _enum.Enum):
    INTERNAL = "internal"
    LOCAL = "local"
    EXTERNAL = "external"


@_enum.unique
class UnaryOperator(_StringyEnumMixin, _enum.Enum):
    """The operator of a unary expression."""

    UNSET = "UNSET"
    """The unary operator is not initialized."""
    NOT = "NOT"
    """The unary operator refers to a NOT operation."""
    POS = "POS"
    """The unary operator refers to a position operation."""
    VAL = "VAL"
    """The unary operator refers to a value operation."""
    SUC = "SUC"
    """The unary operator refers to a successor operation."""
    PRE = "PRE"
    """The unary operator refers to a predecessor operation."""


@_enum.unique
class UnionKind(_StringyEnumMixin, _enum.Enum):
    UNION = "UNION"
    VARIANT = "VARIANT"


@_enum.unique
class VisibilityKind(_StringyEnumMixin, _enum.Enum):
    """The possibilities regarding the visibility of a feature of a class."""

    UNSET = "UNSET"
    """Visibility is not specified."""
    PUBLIC = "PUBLIC"
    """The feature offers public access."""
    PROTECTED = "PROTECTED"
    """The feature offers visibility only to children of the class."""
    PRIVATE = "PRIVATE"
    """The feature is only visible/accessible from the class itself."""
    PACKAGE = "PACKAGE"
    """The feature is accessible from any element within the same package."""


if not t.TYPE_CHECKING:
    __replaced_names = {
        "ExchangeItemType": ExchangeMechanism,
        "Kind": PhysicalComponentKind,
        "Nature": PhysicalComponentNature,
        "FPortDir": OrientationPortKind,
    }

    def __getattr__(name):
        if replacement := __replaced_names.get(name):
            import warnings

            warnings.warn(
                f"{name} is deprecated; use {replacement.__name__} instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return replacement
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
